from joatmon import context
from joatmon.plugin.localizer.core import Localizer
from joatmon.utility import (
    new_object_id,
    to_list_async
)


class DatabaseLocalizer(Localizer):
    def __init__(self, database, cls):
        self.database = database
        self.cls = cls

    async def localize(self, language, keys):
        database = context.get_value(self.database)

        db_resources = await to_list_async(database.read(self.cls, {'key': {'$in': keys}}))
        found_keys = list(map(lambda x: x.key, db_resources))
        not_found_keys = list(filter(lambda x: x not in found_keys, keys))

        for not_found_key in not_found_keys:
            r = {'object_id': new_object_id(), 'key': not_found_key, language: not_found_key}

            await database.insert(self.cls, r)

        for found_key in found_keys:
            db_resource = list(filter(lambda x: x.key == found_key, db_resources))[0]
            if getattr(db_resource, language, None) is None:
                setattr(db_resource, language, db_resource.key)
                await database.update(self.cls, {'object_id': db_resource.object_id}, db_resource)

        db_resources = await to_list_async(database.read(self.cls, {'key': {'$in': keys}}))
        return db_resources
        # return list(map(lambda x: x[language], db_resources))



from __future__ import unicode_literals

from datetime import (
    timedelta
)

import requests

API_URL = 'http://en.wikipedia.org/w/api.php'
RATE_LIMIT = True
RATE_LIMIT_MIN_WAIT = None
RATE_LIMIT_LAST_CALL = timedelta(milliseconds=900)
USER_AGENT = 'wikipedia (https://github.com/goldsmith/Wikipedia/)'


def set_lang(prefix):
    global API_URL
    API_URL = 'http://' + prefix.lower() + '.wikipedia.org/w/api.php'


def summary(title, sentences=1):
    query_params = {'prop': 'info|pageprops', 'inprop': 'url', 'ppprop': 'disambiguation', 'redirects': '', 'titles': title}

    request = _wiki_request(query_params)

    query = request['query']
    pageid = list(query['pages'].keys())[0]
    title = query['pages'][pageid]['title']

    query_params = {'prop': 'extracts', 'explaintext': '', 'titles': title, 'exsentences': sentences}

    request = _wiki_request(query_params)
    return request['query']['pages'][pageid]['extract']


def languages():
    response = _wiki_request({'meta': 'siteinfo', 'siprop': 'languages'})
    return {lang['code']: lang['*'] for lang in response['query']['languages']}


def _wiki_request(params):
    global RATE_LIMIT_LAST_CALL
    global USER_AGENT

    params['format'] = 'json'
    if 'action' not in params:
        params['action'] = 'query'

    r = requests.get(API_URL, params=params, headers={'User-Agent': USER_AGENT})

    return r.json()



def register(*decorators):
    def register_wrapper(func):
        for deco in decorators[::-1]:
            func = deco(func)
        func._decorators = decorators
        return func

    return register_wrapper



import functools
import inspect
import os
import sys
import time
import traceback

from joatmon import context
from joatmon.utility import (
    get_function_args,
    get_function_kwargs,
    to_enumerable
)


def log(logger, on_begin=None, on_success=None, on_error=None):
    # need to stop writing self values
    async def _on_begin(f, *args, **kwargs):
        logger_value = context.get_value(logger)

        await logger_value.info(
            {
                'function': f.__qualname__,
                'module': f.__module__,
                'args': to_enumerable(get_function_args(f, *args), string=True),
                'kwargs': to_enumerable(get_function_kwargs(f, **kwargs), string=True)
            }
        )

    async def _on_success(f, t, result, *args, **kwargs):
        logger_value = context.get_value(logger)

        await logger_value.info(
            {
                'timed': t,
                'result': to_enumerable(result, string=True),
                'function': f.__qualname__,
                'module': f.__module__,
                'args': to_enumerable(get_function_args(f, *args), string=True),
                'kwargs': to_enumerable(get_function_kwargs(f, **kwargs), string=True)
            }
        )

    async def _on_error(f, exception, *args, **kwargs):
        logger_value = context.get_value(logger)

        exc_type, exc_obj, exc_trace = sys.exc_info()

        await logger_value.error(
            {
                'exception': {
                    'line': exc_trace.tb_lineno,
                    'file': os.path.split(exc_trace.tb_frame.f_code.co_filename)[1],
                    'message': str(exception),
                    'trace': traceback.format_exc()
                },
                'function': f.__qualname__,
                'module': f.__module__,
                'args': to_enumerable(get_function_args(f, *args), string=True),
                'kwargs': to_enumerable(get_function_kwargs(f, **kwargs), string=True)
            }
        )

    on_begin = on_begin or _on_begin
    on_success = on_success or _on_success
    on_error = on_error or _on_error

    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            await on_begin(func, *args, **kwargs)
            try:
                begin = time.time()
                result = await func(*args, **kwargs)
                end = time.time()

                await on_success(func, end - begin, result, *args, **kwargs)

                return result
            except Exception as ex:
                await on_error(func, ex, *args, **kwargs)
                raise ex

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator



import numpy as np
from PIL import Image


class RLProcessor:
    def __init__(self):
        super(RLProcessor, self).__init__()

    @staticmethod
    def process_batch(batch):
        states = []
        actions = []
        rewards = []
        next_states = []
        terminals = []
        for state, action, reward, next_state, terminal in batch:
            states.extend(state)
            actions.append(action)
            rewards.append(reward)
            next_states.extend(next_state)
            terminals.append(terminal)
        states = np.asarray(states).astype('float32')
        actions = np.asarray(actions).astype('float32')
        rewards = np.asarray(rewards).astype('float32')
        next_states = np.asarray(next_states).astype('float32')
        terminals = np.asarray(terminals).astype('float32')
        return states, actions, rewards, next_states, terminals

    @staticmethod
    def process_state(state):
        if state is None:
            return

        state = Image.fromarray(state, 'RGB')
        state = state.resize((84, 84))
        state = np.array(state)
        state = np.expand_dims(state, 0)
        state = np.transpose(state, (0, 3, 1, 2))
        state = state.astype('uint8')
        return state



from __future__ import print_function

import subprocess

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "run",
            "description": "a function for user to run an executable",
            "parameters": {
                "type": "object",
                "properties": {
                    "executable": {
                        "type": "string",
                        "description": "executable to run"
                    }
                },
                "required": ["executable"]
            }
        }

    def run(self):
        executable = self.kwargs.get('executable', '') or self.api.input('what do you want to run')

        subprocess.run(['python.exe', executable])

        if not self.stop_event.is_set():
            self.stop_event.set()



import numpy as np


class CorePolicy(object):
    """
    Abstract base class for all implemented policy.

    Do not use this abstract base class directly but
    instead use one of the concrete policy implemented.

    To implement your own policy, you have to implement the following methods:

    - `decay`
    - `use_network`
    """

    def __init__(self):
        super(CorePolicy, self).__init__()

    def reset(self):
        """
        reset
        """
        raise NotImplementedError

    def decay(self):
        """
        Decaying the epsilon / sigma value of the policy.
        """
        raise NotImplementedError

    def use_network(self):
        """
        Sample an experience replay batch with size.

        # Returns
            use (bool): Boolean value for using the nn.
        """
        raise NotImplementedError


class GreedyQPolicy(CorePolicy):
    def __init__(self):
        super().__init__()

    def reset(self):
        pass

    def decay(self):
        pass

    def use_network(self):
        return True


class EpsilonGreedyPolicy(CorePolicy):
    """
    Epsilon Greedy

    # Arguments
        max_value (float): .
        min_value (float): .
        decay_steps (int): .
    """

    def __init__(self, max_value=1.0, min_value=0.0, decay_steps=1):
        super().__init__()

        self.epsilon = max_value
        self.min_value = min_value
        self.epsilon_decay = (max_value - min_value) / decay_steps

    def reset(self):
        pass

    def decay(self):
        self.epsilon -= self.epsilon_decay
        self.epsilon = max(self.epsilon, self.min_value)

    def use_network(self):
        return np.random.uniform() >= self.epsilon



import ctypes
import sys

from joatmon.system.decorators import auto_pause
from joatmon.system.hid.screen import (
    cursor,
    resolution
)

__all__ = ['Mouse']

if sys.platform != 'win32':
    raise Exception('The mouse module should only be used on a Windows system.')


@auto_pause(duration=0.05)
def _send_mouse_event(x, y, event):
    height, width = resolution()
    converted_x = 65536 * x // width + 1
    converted_y = 65536 * y // height + 1
    ctypes.windll.user32.mouse_event(event, ctypes.c_long(converted_x), ctypes.c_long(converted_y), 0, 0)


class Mouse:
    MOUSE_DOWN = 0x0000
    MOUSE_UP = 0x0001
    MOUSE_LEFT = 0x0002
    MOUSE_RIGHT = 0x0008
    MOUSE_MIDDLE = 0x0020

    MIN_X = None
    MAX_X = None
    MIN_Y = None
    MAX_Y = None

    @staticmethod
    def restrict(min_x, max_x, min_y, max_y):
        Mouse.min_x = min_x
        Mouse.max_x = max_x
        Mouse.min_y = min_y
        Mouse.max_y = max_y

    @auto_pause(duration=0.05)
    def move_to(self, x=None, y=None):
        # needs a time to move
        if x is None:
            _, x = cursor()
        if y is None:
            y, _ = cursor()

        height, width = resolution()

        min_x = 0
        if Mouse.MIN_X is not None:
            min_x = Mouse.MIN_X

        min_y = 0
        if Mouse.MIN_Y is not None:
            min_y = Mouse.MIN_Y

        max_x = width
        if Mouse.MAX_X is not None:
            max_x = Mouse.MAX_X

        max_y = height
        if Mouse.MAX_Y is not None:
            max_y = Mouse.MAX_Y

        x = max(0, min(x, width - 1))
        y = max(0, min(y, height - 1))

        x = max(min_x, min(x, max_x))
        y = max(min_y, min(y, max_y))

        x = int(x)
        y = int(y)

        ctypes.windll.user32.SetCursorPos(x, y)

    @staticmethod
    def mouse_down(x=None, y=None, button=None):
        if button is None:
            button = Mouse.MOUSE_LEFT

        Mouse.move_to(x, y)

        y, x = cursor()
        _send_mouse_event(x, y, button << Mouse.MOUSE_DOWN)

    @staticmethod
    def mouse_up(x=None, y=None, button=None):
        if button is None:
            button = Mouse.MOUSE_LEFT

        Mouse.move_to(x, y)

        y, x = cursor()
        _send_mouse_event(x, y, button << Mouse.MOUSE_UP)

    @staticmethod
    def click(x=None, y=None, button=None):
        if button is None:
            button = Mouse.MOUSE_LEFT

        Mouse.move_to(x, y)

        Mouse.mouse_down(button=button)
        Mouse.mouse_up(button=button)



import io
import logging
from asyncio import (
    StreamReader,
    StreamWriter
)

from websockets.exceptions import ConnectionClosed
from websockets.protocol import WebSocketCommonProtocol


class ReaderAdapter:
    async def read(self, n=-1) -> bytes:
        ...

    def feed_eof(self):
        ...


class WriterAdapter:
    def write(self, data):
        ...

    async def drain(self):
        ...

    def get_peer_info(self):
        ...

    async def close(self):
        ...


class WebSocketsReader(ReaderAdapter):
    def __init__(self, protocol: WebSocketCommonProtocol):
        self._protocol = protocol
        self._stream = io.BytesIO(b'')

    async def read(self, n=-1) -> bytes:
        await self._feed_buffer(n)
        data = self._stream.read(n)
        return data

    async def _feed_buffer(self, n=1):
        buffer = bytearray(self._stream.read())
        while len(buffer) < n:
            try:
                message = await self._protocol.recv()
            except ConnectionClosed:
                message = None
            if message is None:
                break
            if not isinstance(message, bytes):
                raise TypeError("message must be bytes")
            buffer.extend(message)
        self._stream = io.BytesIO(buffer)


class WebSocketsWriter(WriterAdapter):
    def __init__(self, protocol: WebSocketCommonProtocol):
        self._protocol = protocol
        self._stream = io.BytesIO(b'')

    def write(self, data):
        self._stream.write(data)

    async def drain(self):
        data = self._stream.getvalue()
        if len(data):
            await self._protocol.send(data)
        self._stream = io.BytesIO(b'')

    def get_peer_info(self):
        return self._protocol.remote_address

    async def close(self):
        await self._protocol.close()


class StreamReaderAdapter(ReaderAdapter):
    def __init__(self, reader: StreamReader):
        self._reader = reader

    async def read(self, n=-1) -> bytes:
        if n == -1:
            data = await self._reader.read(n)
        else:
            data = await self._reader.readexactly(n)
        return data

    def feed_eof(self):
        return self._reader.feed_eof()


class StreamWriterAdapter(WriterAdapter):
    def __init__(self, writer: StreamWriter):
        self.logger = logging.getLogger(__name__)
        self._writer = writer
        self.is_closed = False

    def write(self, data):
        if self.is_closed:
            return
        self._writer.write(data)

    async def drain(self):
        if self.is_closed:
            return
        await self._writer.drain()

    def get_peer_info(self):
        extra_info = self._writer.get_extra_info('peername')
        return extra_info[0], extra_info[1]

    async def close(self):
        if self.is_closed:
            return

        self.is_closed = True
        await self._writer.drain()
        if self._writer.can_write_eof():
            self._writer.write_eof()
        self._writer.close()

        try:
            await self._writer.wait_closed()
        except AttributeError:
            pass



from collections import deque


def inorder(root):
    stack = deque()
    curr = root

    while stack or curr:
        if curr:
            stack.append(curr)
            curr = curr.left
        else:
            curr = stack.pop()
            yield curr.data
            curr = curr.right


def inorder_r(root):
    if root is None:
        return

    inorder_r(root.left)
    yield root.data
    inorder_r(root.right)



from joatmon.nn import functional as f
from joatmon.nn.core import Module

__all__ = ['Softmax']


class Softmax(Module):
    def __init__(self):
        super(Softmax, self).__init__()

    def forward(self, inp):
        return f.softmax(
            inp=inp
        )



import jwt

from joatmon.plugin.auth.core import Auth


class JWTAuth(Auth):
    def __init__(self, secret: str):
        self.secret = secret

    async def authenticate(self, issuer, audience, expire_at, **kwargs):
        kwargs.update({'exp': expire_at, 'iss': issuer, 'aud': audience})

        token = jwt.encode(
            payload=kwargs,
            key=self.secret,
            algorithm='HS256',
        )
        return token

    async def authorize(self, token, issuer, audience):
        try:
            decoded = jwt.decode(token, self.secret, issuer=issuer, audience=audience, algorithms='HS256')
        except jwt.DecodeError:
            raise ValueError('token_decode_error')
        except jwt.ExpiredSignatureError:
            raise ValueError('token_expired')
        except ValueError:
            raise ValueError('token_decode_error')

        return decoded



import ctypes
import os
import platform
import struct
from ctypes import wintypes

from joatmon.system.memory.core import (
    CreateToolhelp32Snapshot,
    DACL_SECURITY_INFORMATION,
    IsWow64Process,
    MEM_FREE,
    MEM_RESERVE,
    MemoryBasicInformation,
    Module32First,
    Module32Next,
    ModuleEntry32,
    NtWow64ReadVirtualMemory64,
    PAGE_EXECUTE_READWRITE,
    PAGE_GUARD,
    PAGE_NOCACHE,
    PAGE_WRITECOMBINE,
    ReadProcessMemory,
    SecurityDescriptor,
    SystemInfo,
    Th32csClass,
    type_unpack,
    UNPROTECTED_DACL_SECURITY_INFORMATION,
    VirtualQueryEx
)


class ProcessException(Exception):
    pass


class BaseProcess(object):
    def __init__(self, *args, **kwargs):
        self.h_process = None
        self.pid = None
        self.is_process_open = False
        self.buffer = None
        self.buffer_len = 0

    def __del__(self):
        self.close()

    def close(self):
        pass

    def iter_region(self, *args, **kwargs):
        raise NotImplementedError

    def write_bytes(self, address, data):
        raise NotImplementedError

    def read_bytes(self, address, size=4):
        raise NotImplementedError

    @staticmethod
    def get_symbolic_name(a):
        return '0x%08X' % int(a)

    def read(self, address, of_type='uint', max_length=50, errors='raise'):
        if of_type == 's' or of_type == 'string':
            byte_array = self.read_bytes(int(address), size=max_length)
            new_string = ''
            for character in byte_array:
                if character == '\x00':
                    return new_string
                new_string += character
            if errors == 'ignore':
                return new_string
            raise ProcessException('string > maxlen')
        else:
            if of_type == 'bytes' or of_type == 'b':
                return self.read_bytes(int(address), size=max_length)
            struct_type, struct_length = type_unpack(of_type)
            return struct.unpack(struct_type, self.read_bytes(int(address), size=struct_length))[0]

    def write(self, address, data, of_type='uint'):
        if of_type != 'bytes':
            struct_type, struct_length = type_unpack(of_type)
            return self.write_bytes(int(address), struct.pack(struct_type, data))
        else:
            return self.write_bytes(int(address), data)


class Process(BaseProcess):
    def __init__(self, pid=None, name=None, debug=True):
        super(Process, self).__init__()
        if pid:
            self._open(int(pid), debug=debug)
        elif name:
            self._open_from_name(name, debug=debug)
        else:
            raise ValueError("You need to instanciate process with at least a name or a pid")

        if self.is_64():
            si = self.get_native_system_info()
            self.max_address = si.lpMaximumApplicationAddress
        else:
            si = self.get_system_info()
            self.max_address = 2147418111

        self.min_address = si.lpMinimumApplicationAddress

    def __del__(self):
        self.close()

    def is_64(self):
        if "64" not in platform.machine():
            return False
        is_wow_64 = ctypes.c_bool(False)
        if IsWow64Process is None:
            return False
        if not IsWow64Process(self.h_process, ctypes.byref(is_wow_64)):
            raise ctypes.WinError()
        return not is_wow_64.value

    @staticmethod
    def list():
        processes = []
        arr = ctypes.c_ulong * 256
        l_pid_process = arr()
        cb = ctypes.sizeof(l_pid_process)
        cb_needed = ctypes.c_ulong()
        h_module = ctypes.c_ulong()
        count = ctypes.c_ulong()
        modname = ctypes.create_string_buffer(100)
        process_query_information = 0x0400
        process_vm_read = 0x0010

        ctypes.windll.psapi.EnumProcesses(ctypes.byref(l_pid_process), cb, ctypes.byref(cb_needed))
        n_returned = cb_needed.value / ctypes.sizeof(ctypes.c_ulong())

        pid_process = [i for i in l_pid_process][:int(n_returned)]
        for pid in pid_process:
            proc = {
                "pid": int(pid)
            }
            h_process = ctypes.windll.kernel32.OpenProcess(process_query_information | process_vm_read, False, pid)
            if h_process:
                ctypes.windll.psapi.EnumProcessModules(
                    h_process,
                    ctypes.byref(h_module),
                    ctypes.sizeof(h_module),
                    ctypes.byref(count)
                )
                ctypes.windll.psapi.GetModuleBaseNameA(h_process, h_module.value, modname, ctypes.sizeof(modname))
                proc["name"] = modname.value.decode('utf-8')
                ctypes.windll.kernel32.CloseHandle(h_process)
            processes.append(proc)
        return processes

    @staticmethod
    def processes_from_name(process_name):
        processes = []
        for process in Process.list():
            if (
                    process_name == process.get("name", None) or
                    (process.get("name", "").lower().endswith(".exe") and
                     process.get("name", "")[:-4] == process_name)
            ):
                processes.append(process)

        if len(processes) > 0:
            return processes

    @staticmethod
    def name_from_process(dw_process_id):
        process_list = Process.list()
        for process in process_list:
            if process.get('pid', -1) == dw_process_id:
                return process.get("name", None)

        return False

    def _open(self, dw_process_id, debug=False):
        if debug:
            ppsid_owner = wintypes.DWORD()
            ppsid_group = wintypes.DWORD()
            pp_dacl = wintypes.DWORD()
            pp_sacl = wintypes.DWORD()
            pp_security_descriptor = SecurityDescriptor()

            process = ctypes.windll.kernel32.OpenProcess(262144, 0, dw_process_id)
            ctypes.windll.advapi32.GetSecurityInfo(
                ctypes.windll.kernel32.GetCurrentProcess(),
                6,
                0,
                ctypes.byref(ppsid_owner),
                ctypes.byref(ppsid_group),
                ctypes.byref(pp_dacl),
                ctypes.byref(pp_sacl),
                ctypes.byref(pp_security_descriptor)
            )
            ctypes.windll.advapi32.SetSecurityInfo(
                process,
                6,
                DACL_SECURITY_INFORMATION | UNPROTECTED_DACL_SECURITY_INFORMATION,
                None,
                None,
                pp_security_descriptor.dacl,
                pp_security_descriptor.group
            )
            ctypes.windll.kernel32.CloseHandle(process)
        self.h_process = ctypes.windll.kernel32.OpenProcess(2035711, 0, dw_process_id)
        if self.h_process is not None:
            self.is_process_open = True
            self.pid = dw_process_id
            return True
        return False

    def close(self):
        if self.h_process is not None:
            ret = ctypes.windll.kernel32.CloseHandle(self.h_process) == 1
            if ret:
                self.h_process = None
                self.pid = None
                self.is_process_open = False
            return ret
        return False

    def _open_from_name(self, process_name, debug=False):
        processes = self.processes_from_name(process_name)
        # need to remove the self process
        processes = list(filter(lambda process: process.get('pid', -1) != os.getpid(), processes))
        if not processes:
            raise ProcessException("can't get pid from name %s" % process_name)
        elif len(processes) > 1:
            raise ValueError(
                "There is multiple processes with name %s. Please select a process from its pid instead" % process_name
            )

        if debug:
            self._open(processes[0]["pid"], debug=True)
        else:
            self._open(processes[0]["pid"], debug=False)

    @staticmethod
    def get_system_info():
        si = SystemInfo()
        ctypes.windll.kernel32.GetSystemInfo(ctypes.byref(si))
        return si

    @staticmethod
    def get_native_system_info():
        si = SystemInfo()
        ctypes.windll.kernel32.GetNativeSystemInfo(ctypes.byref(si))
        return si

    def virtual_query_ex(self, lp_address):
        mbi = MemoryBasicInformation()
        if not VirtualQueryEx(self.h_process, lp_address, ctypes.byref(mbi), ctypes.sizeof(mbi)):
            raise ProcessException('Error VirtualQueryEx: 0x%08X' % lp_address)
        return mbi

    def virtual_protect_ex(self, base_address, size, protection):
        old_protect = ctypes.c_ulong(0)
        if not ctypes.windll.kernel32.virtual_protect_ex(self.h_process, base_address, size, protection, ctypes.byref(old_protect)):
            raise ProcessException('Error: VirtualProtectEx(%08X, %d, %08X)' % (base_address, size, protection))
        return old_protect.value

    def iter_region(self, start_offset=None, end_offset=None, protect=None, optimizations=None):
        _offset_start = start_offset or self.min_address
        _offset_end = end_offset or self.max_address

        while True:
            if _offset_start >= _offset_end:
                break
            mbi = self.virtual_query_ex(_offset_start)
            _offset_start = mbi.BaseAddress
            _chunk = mbi.RegionSize
            _protect = mbi.Protect
            _state = mbi.State

            if _state & MEM_FREE or _state & MEM_RESERVE:
                _offset_start += _chunk
                continue
            if protect:
                if not _protect & protect or _protect & PAGE_NOCACHE or _protect & PAGE_WRITECOMBINE or _protect & PAGE_GUARD:
                    _offset_start += _chunk
                    continue
            yield _offset_start, _chunk
            _offset_start += _chunk

    def write_bytes(self, address, data):
        address = int(address)
        if not self.is_process_open:
            raise ProcessException("Can't write_bytes(%s, %s), process %s is not open" % (address, data, self.pid))
        buffer = ctypes.create_string_buffer(data)
        size_writen = ctypes.c_size_t(0)
        buffer_size = ctypes.sizeof(buffer) - 1
        _address = address
        _length = buffer_size + 1
        try:
            old_protect = self.virtual_protect_ex(_address, _length, PAGE_EXECUTE_READWRITE)
        except Exception as ex1:
            print(str(ex1))

        res = ctypes.windll.kernel32.WriteProcessMemory(
            self.h_process,
            address,
            buffer,
            buffer_size,
            ctypes.byref(size_writen)
        )
        try:
            self.virtual_protect_ex(_address, _length, old_protect)
        except Exception as ex2:
            print(str(ex2))

        return res

    def read_bytes(self, address, size=4, use_nt_wow64_read_virtual_memory64=False):
        if use_nt_wow64_read_virtual_memory64:
            if NtWow64ReadVirtualMemory64 is None:
                raise WindowsError("NtWow64ReadVirtualMemory64 is not available from a 64bit process")
            rpm = NtWow64ReadVirtualMemory64
        else:
            rpm = ReadProcessMemory

        address = int(address)
        buffer = ctypes.create_string_buffer(size)
        bytes_read = ctypes.c_size_t(0)
        data = b''
        length = size
        while length:
            if rpm(self.h_process, address, buffer, size, ctypes.byref(bytes_read)) or (use_nt_wow64_read_virtual_memory64 and ctypes.GetLastError() == 0):
                if bytes_read.value:
                    data += buffer.raw[:bytes_read.value]
                    length -= bytes_read.value
                    address += bytes_read.value
                if not len(data):
                    raise ProcessException(
                        'Error %s in ReadProcessMemory(%08x, %d, read=%d)' % (
                            ctypes.GetLastError(),
                            address,
                            length,
                            bytes_read.value
                        )
                    )
                return data
            else:
                if ctypes.GetLastError() == 299:
                    data += buffer.raw[:bytes_read.value]
                    return data
                raise ctypes.WinError()
        return data

    def list_modules(self):
        if self.pid is not None:
            h_module_snap = CreateToolhelp32Snapshot(Th32csClass.SNAPMODULE, self.pid)
            if h_module_snap is not None:
                module_entry = ModuleEntry32()
                module_entry.dwSize = ctypes.sizeof(module_entry)
                success = Module32First(h_module_snap, ctypes.byref(module_entry))
                while success:
                    if module_entry.th32ProcessID == self.pid:
                        yield module_entry
                    success = Module32Next(h_module_snap, ctypes.byref(module_entry))

                ctypes.windll.kernel32.CloseHandle(h_module_snap)

    def get_symbolic_name(self, address):
        for m in self.list_modules():
            if int(ctypes.addressof(m.modBaseAddr.contents)) <= int(address) < int(ctypes.addressof(m.modBaseAddr.contents) + m.modBaseSize):
                return '%s+0x%08X' % (m.szModule, int(address) - ctypes.addressof(m.modBaseAddr.contents))

        return '0x%08X' % int(address)

    def has_module(self, module):
        if module[-4:] != '.dll':
            module += '.dll'
        module_list = self.list_modules()
        for m in module_list:
            if module in m.szExePath.split('\\'):
                return True
        return False



import pygame

from joatmon.event import Event


class Element:
    def __init__(self, id_):
        self._id = id_

        self._original_w = 1
        self._original_h = 1

        self._x = 0
        self._y = 0

        self._width = 1
        self._height = 1

        self._color = (255, 255, 255)

        self._border = {'color': (0, 0, 0), 'thickness': 1, 'shadow': 1, 'radius': 1}
        self._rect = None

        self._surface = None
        self._elements = []
        self._layout = 'vertical'

        self._events = {'click': Event()}

    def layout(self, layout):
        self._layout = layout

        return self

    def add(self, element):
        self._elements.append(element)
        self.update()

        return self

    def style(self, style):
        self._width = style.get('width', self._width)
        self._height = style.get('height', self._height)

        self._original_w = style.get('width', self._original_w)
        self._original_h = style.get('height', self._original_h)

        self._rect = pygame.Rect(self._x, self._y, self._width, self._height)
        # self._surface = pygame.Surface((self._width, self._height))

        self._color = style.get('color', self._color)

        self._border = style.get('border', self._border)

        return self

    def on(self, event, func):
        _event = self._events.get(event, None)
        if _event:
            _event += func

        return self

    def update(self):
        self._rect = pygame.Rect(self._x, self._y, self._width, self._height)

        if self._layout == 'vertical':
            h_weights = list(map(lambda x: x._original_h, self._elements))
            w_weights = list(map(lambda x: 1, self._elements))
        else:
            h_weights = list(map(lambda x: 1, self._elements))
            w_weights = list(map(lambda x: x._original_w, self._elements))

        for idx, element in enumerate(self._elements):
            if self._layout == 'vertical':
                element._width = self._width
                element._height = h_weights[idx] * (self._height / sum(h_weights))
            else:
                element._width = w_weights[idx] * (self._width / sum(w_weights))
                element._height = self._height

        current_x = self._x
        current_y = self._y
        for element in self._elements:
            element._x = current_x
            element._y = current_y

            if self._layout == 'vertical':
                current_y += element._height
            else:
                current_x += element._width

        for element in self._elements:
            element.update()

    def draw(self, screen):
        border_color = self._border.get('color', (0, 0, 0))
        thickness = self._border.get('thickness', 1)
        radius = self._border.get('radius', 1)
        shadow = self._border.get('shadow', 3)

        if self._border:
            border_rect = self._rect.copy()
            border_rect.x -= thickness
            border_rect.y -= thickness
            border_rect.width += 2 * thickness
            border_rect.height += 2 * thickness

            if shadow:
                shadow_rect = border_rect.copy()
                shadow_rect.x += shadow
                shadow_rect.y += shadow
                pygame.draw.rect(screen, pygame.Color('gray'), shadow_rect, border_radius=radius)  # shadow color

                pygame.draw.rect(screen, self._color, border_rect, border_radius=radius)
            pygame.draw.rect(screen, border_color, border_rect, thickness, border_radius=radius)

        pygame.draw.rect(screen, self._color, self._rect)

        for element in self._elements:
            element.draw(screen)

    def handle(self, event):
        for element in self._elements:
            element.handle(event)

        if event.type == pygame.MOUSEBUTTONDOWN and self._rect.collidepoint(event.pos):
            _event = self._events.get('click', None)
            if _event:
                _event.fire(self)



import asyncio
from asyncio import Queue
from collections import OrderedDict

from transitions import Machine

from .errors import HZMQTTException
from .mqtt.packet import PublishPacket

OUTGOING = 0
INCOMING = 1


class ApplicationMessage:
    __slots__ = ('packet_id', 'topic', 'qos', 'data', 'retain', 'publish_packet', 'puback_packet', 'pubrec_packet', 'pubrel_packet', 'pubcomp_packet')

    def __init__(self, packet_id, topic, qos, data, retain):
        self.packet_id = packet_id
        self.topic = topic
        self.qos = qos
        self.data = data
        self.retain = retain
        self.publish_packet = None
        self.puback_packet = None
        self.pubrec_packet = None
        self.pubrel_packet = None
        self.pubcomp_packet = None

    def build_publish_packet(self, dup=False):
        return PublishPacket.build(self.topic, self.data, self.packet_id, dup, self.qos, self.retain)

    def __eq__(self, other):
        return self.packet_id == other.packet_id


class IncomingApplicationMessage(ApplicationMessage):
    __slots__ = ('direction',)

    def __init__(self, packet_id, topic, qos, data, retain):
        super().__init__(packet_id, topic, qos, data, retain)
        self.direction = INCOMING


class OutgoingApplicationMessage(ApplicationMessage):
    __slots__ = ('direction',)

    def __init__(self, packet_id, topic, qos, data, retain):
        super().__init__(packet_id, topic, qos, data, retain)
        self.direction = OUTGOING


class Session:
    states = ['new', 'connected', 'disconnected']

    def __init__(self, loop=None):
        self._init_states()
        self.remote_address = None
        self.remote_port = None
        self.client_id = None
        self.clean_session = None
        self.will_flag = False
        self.will_message = None
        self.will_qos = None
        self.will_retain = None
        self.will_topic = None
        self.keep_alive = 0
        self.publish_retry_delay = 0
        self.broker_uri = None
        self.username = None
        self.password = None
        self.cafile = None
        self.capath = None
        self.cadata = None
        self._packet_id = 0
        self.parent = 0
        if loop is not None:
            self._loop = loop
        else:
            self._loop = asyncio.get_event_loop()

        self.inflight_out = OrderedDict()
        self.inflight_in = OrderedDict()
        self.retained_messages = Queue(loop=self._loop)
        self.delivered_message_queue = Queue(loop=self._loop)

    def _init_states(self):
        self.transitions = Machine(states=Session.states, initial='new')
        self.transitions.add_transition(trigger='connect', source='new', dest='connected')
        self.transitions.add_transition(trigger='connect', source='disconnected', dest='connected')
        self.transitions.add_transition(trigger='disconnect', source='connected', dest='disconnected')
        self.transitions.add_transition(trigger='disconnect', source='new', dest='disconnected')
        self.transitions.add_transition(trigger='disconnect', source='disconnected', dest='disconnected')

    @property
    def next_packet_id(self):
        self._packet_id += 1
        if self._packet_id > 65535:
            self._packet_id = 1
        while self._packet_id in self.inflight_in or self._packet_id in self.inflight_out:
            self._packet_id += 1
            if self._packet_id > 65535:
                raise HZMQTTException("More than 65525 messages pending. No free packet ID")

        return self._packet_id

    @property
    def inflight_in_count(self):
        return len(self.inflight_in)

    @property
    def inflight_out_count(self):
        return len(self.inflight_out)

    @property
    def retained_messages_count(self):
        return self.retained_messages.qsize()

    def __repr__(self):
        return type(self).__name__ + '(clientId={0}, state={1})'.format(self.client_id, self.transitions.state)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['retained_messages']
        del state['delivered_message_queue']
        return state

    def __setstate(self, state):
        self.__dict__.update(state)
        self.retained_messages = Queue()
        self.delivered_message_queue = Queue()

    def __eq__(self, other):
        return self.client_id == other.client_id



import sys

import cv2
import numpy
import numpy as np

from joatmon.system.hid.screen import grab

__all__ = ['watermark', 'crop', 'show', 'save', 'mask', 'match_template']

if sys.platform != 'win32':
    raise Exception('The image module should only be used on a Windows system.')


def watermark(image=None, mark=None, region=None, alpha=0.2) -> numpy.ndarray:
    if mark is None:
        raise Exception

    if image is None:
        image = grab()

    if region:
        left, top, right, bottom = region
    else:
        raise Exception

    image_copy = image.copy()
    image_copy = numpy.dstack([image_copy, numpy.ones(image_copy.shape[:2], dtype='uint8') * 255])

    overlay = numpy.zeros(image_copy.shape[:2] + (4,), dtype='uint8')
    overlay[top:bottom, left:right] = mark

    output = image_copy.copy()
    cv2.addWeighted(overlay, alpha, output, 1.0, 0.0, output)
    return cv2.cvtColor(output, cv2.COLOR_BGRA2BGR)


def crop(image=None, region=None) -> numpy.ndarray:
    if image is None:
        image = grab()

    if region:
        left, top, right, bottom = region
    else:
        left, top, (right, bottom) = 0, 0, *image.size

    return image[top: bottom, left: right]


def show(image, title) -> None:
    cv2.imshow(title, image)
    cv2.waitKey(1)


def save(image, path) -> None:
    cv2.imwrite(path, image)


def mask(image=None, r1=(0, 255), g1=(0, 255), b1=(0, 255), r2=None, g2=None, b2=None):
    if image is None:
        image = grab()

    hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

    lower1 = np.array([r1[0], g1[0], b1[0]])
    upper1 = np.array([r1[1], g1[1], b1[1]])
    mask1 = cv2.inRange(hsv, lower1, upper1)

    masked = mask1

    if r2 is not None and g2 is not None and b2 is not None:
        lower2 = np.array([r2[0], g2[0], b2[0]])
        upper2 = np.array([r2[1], g2[1], b2[1]])
        mask2 = cv2.inRange(hsv, lower2, upper2)

        masked += mask2

    copy = image.copy()
    copy[np.where(masked == 0)] = 0
    return copy


def match_template(image=None, template=None, threshold=0.8) -> list:
    if template is None:
        raise Exception

    if image is None:
        image = grab()

    res = cv2.matchTemplate(cv2.cvtColor(image, cv2.COLOR_BGR2GRAY), cv2.cvtColor(template, cv2.COLOR_BGR2GRAY), cv2.TM_CCOEFF_NORMED)
    loc = numpy.where(res >= threshold)
    return list(zip(*loc[::-1]))



import json

from joatmon.plugin.logger.core import LoggerPlugin


class TerminalLogger(LoggerPlugin):
    def __init__(self, level: str, language, ip):
        super(TerminalLogger, self).__init__(level, language, ip)

    async def _write(self, log: dict):
        print(json.dumps(log))



from __future__ import print_function

import os

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "rm",
            "description": "a function for user to remove a file or a folder with its content",
            "parameters": {
                "type": "object",
                "properties": {
                    "path": {
                        "type": "string",
                        "description": "path to remove"
                    }
                },
                "required": ["path"]
            }
        }

    def run(self):
        path = self.kwargs.get('path', '') or self.api.input('which folder/file do you want to remove')

        parent_os_path = self.kwargs.get('parent_os_path', '')
        os_path = self.kwargs.get('os_path', '')

        if os.path.isdir(os.path.join(parent_os_path, os_path[1:], path)):
            os.removedirs(os.path.join(parent_os_path, os_path[1:], path))
        else:
            os.remove(os.path.join(parent_os_path, os_path[1:], path))

        if not self.stop_event.is_set():
            self.stop_event.set()



import functools
import inspect


def retry(times=5):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                print(str(ex))
                return None
            finally:
                return None

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator



from gym.envs.registration import register

from joatmon.game.cube import *
from joatmon.game.puzzle import *
from joatmon.game.sokoban import *
from joatmon.game.tiles import *

register(
    'Sokoban-v0',
    entry_point='joatmon.game.sokoban:SokobanEnv',
    kwargs={
        'xml': 'default.xml',
        'xmls': 'game/assets/sokoban/xmls',
        'sprites': 'game/assets/sokoban/sprites'
    }
)



from joatmon.nn import functional as f
from joatmon.nn.core import Module

__all__ = ['Sigmoid']


class Sigmoid(Module):
    def __init__(self):
        super(Sigmoid, self).__init__()

    def forward(self, inp):
        return f.sigmoid(
            inp=inp
        )



from __future__ import print_function

import webbrowser

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "browser",
            "description": "a function for user to open a browser using a link",
            "parameters": {
                "type": "object",
                "properties": {
                    "url": {
                        "type": "string",
                        "description": "url to open in the browser"
                    }
                },
                "required": ["url"]
            }
        }

    def run(self):
        webbrowser.register('firefox', None, webbrowser.BackgroundBrowser(r'C:\Program Files\Mozilla Firefox\firefox.exe'))
        webbrowser.get('firefox').open_new_tab(self.kwargs.get('url', None))

        if not self.stop_event.is_set():
            self.stop_event.set()



from collections import deque


def bfs(graph, v, discovered):
    q = deque()
    discovered[v] = True
    q.append(v)

    while q:
        v = q.popleft()
        yield v

        for u in graph.adj_list[v]:
            if not discovered[u]:
                discovered[u] = True
                q.append(u)


def bfs_r(graph, q, discovered):
    if not q:
        return

    v = q.popleft()
    yield v

    for u in graph.adj_list[v]:
        if not discovered[u]:
            discovered[u] = True
            q.append(u)

    for x in bfs_r(graph, q, discovered):
        yield x



import datetime
import os
import subprocess


def get_version(version=None):
    if version is None:
        from joatmon import VERSION as v
        version = v
    else:
        assert len(version) == 5
        assert version[3] in ('alpha', 'beta', 'rc', 'final')

    parts = 2 if version[2] == 0 else 3
    main = '.'.join(str(x) for x in version[:parts])

    sub = ''
    if version[3] == 'alpha' and version[4] == 0:
        git_changeset = get_git_changeset()
        if git_changeset:
            sub = '.dev%s' % git_changeset
    elif version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version[3]] + str(version[4])

    return str(main + sub)


def get_git_changeset():
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    git_log = subprocess.Popen(
        'git log --pretty=format:%ct --quiet -1 HEAD',
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, shell=True, cwd=repo_dir, universal_newlines=True
    )
    timestamp = git_log.communicate()[0]
    try:
        timestamp = datetime.datetime.utcfromtimestamp(int(timestamp))
    except ValueError:
        return None
    return timestamp.strftime('%Y%m%d%H%M%S')



import torch
import torch.nn as nn
import torch.nn.functional as F


class DoubleConv(nn.Module):
    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diff_y = x2.size()[2] - x1.size()[2]
        diff_x = x2.size()[3] - x1.size()[3]

        x1 = F.pad(
            x1, [diff_x // 2, diff_x - diff_x // 2,
                 diff_y // 2, diff_y - diff_y // 2]
        )
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)


class UNet(nn.Module):
    def __init__(self, channels, classes, bilinear=False):
        super(UNet, self).__init__()
        self.n_channels = channels
        self.n_classes = classes
        self.bilinear = bilinear

        self.inc = (DoubleConv(channels, 64))
        self.down1 = (Down(64, 128))
        self.down2 = (Down(128, 256))
        self.down3 = (Down(256, 512))
        factor = 2 if bilinear else 1
        self.down4 = (Down(512, 1024 // factor))
        self.up1 = (Up(1024, 512 // factor, bilinear))
        self.up2 = (Up(512, 256 // factor, bilinear))
        self.up3 = (Up(256, 128 // factor, bilinear))
        self.up4 = (Up(128, 64, bilinear))
        self.outc = (OutConv(64, classes))

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        logits = self.outc(x)
        return logits



import json
from io import BytesIO

import openai


class STTAgent:
    def __init__(self):
        super(STTAgent, self).__init__()

        openai.api_key = json.loads(open('iva/iva.json', 'r').read())['config']['openai']['key']

    def transcribe(self, audio):
        wav_data = BytesIO(audio.get_wav_data())
        wav_data.name = "SpeechRecognition_audio.wav"

        transcript = openai.Audio.transcribe('whisper-1', wav_data)
        return transcript["text"]

        # ignore = ['!', '?', ',', '.']
        # result = self.result_queue.get()
        # return ''.join([c.lower() for c in result.lower().strip() if c not in ignore])



import os

import numpy as np
import torch


def load(network, path):
    if path is None or path == '':
        path = os.getcwd()
    if not os.path.exists(path):
        return

    network_path = os.path.join(path, f'network-weights.pth')

    if not os.path.exists(network_path):
        pass
    else:
        weights = torch.load(network_path)

        model_weights = network.state_dict()
        model_keys = list(model_weights.keys())

        for idx in range(len(model_keys)):
            try:
                model_key = model_keys[idx]
                key = model_key.replace('head.', '').replace('body.', '')

                model_weights[model_key] = weights[key]
            except Exception as ex:
                print(str(ex))

        network.load_state_dict(model_weights)


def save(network, path):
    if path is None or path == '':
        path = os.getcwd()
    if not os.path.exists(path):
        os.makedirs(path)

    network_path = os.path.join(path, f'network-weights.pth')

    weights = {}

    model_weights = network.state_dict()
    model_keys = list(model_weights.keys())

    for idx in range(len(model_keys)):
        try:
            model_key = model_keys[idx]
            key = model_key.replace('head.', '').replace('body.', '')

            weights[key] = model_weights[model_key]
        except Exception as ex:
            print(str(ex))

    torch.save(weights, network_path)


def display(values, positions):
    line = ''
    for i in range(len(values)):
        if i > 0:
            line = line[:-1] + ' '
        line += str(values[i])
        line = line[:positions[i]]
        line += ' ' * (positions[i] - len(line))
    print(line)


def easy_range(begin=0, end=None, increment=1):
    counter = begin
    while True:
        if end is not None:
            if counter > end:
                break
        yield counter
        counter += increment


def normalize(array, minimum=0.0, maximum=255.0, dtype='float32'):
    # nanmin and nanmax could be used
    array_minimum = float(np.amin(array))
    array_maximum = float(np.amax(array))
    # print(array_minimum, array_maximum, minimum, maximum)

    return np.asarray((array - array_minimum) * (maximum - minimum) / (array_maximum - array_minimum) + minimum, dtype=dtype)


def range_tensor(end):
    return torch.arange(end).long()


def to_numpy(t):
    return t.cpu().detach().numpy()


def to_tensor(x):
    if isinstance(x, torch.Tensor):
        return x
    x = np.asarray(x, dtype=np.float32)
    x = torch.tensor(x, dtype=torch.float32)
    return x



import copy
import gc
import math
from copy import deepcopy

import numpy as np
from numpy import linalg

from joatmon.ai.utility import easy_range


class DQNTrainer:
    def __init__(self, environment, memory, processor, model, callbacks, test_policy, train_policy, her=False, action_num=4):
        super(DQNTrainer, self).__init__()

        self.environment = environment
        self.memory = memory
        self.processor = processor
        self.model = model
        self.callbacks = callbacks
        self.test_policy = test_policy
        self.train_policy = train_policy
        self.her = her
        self.action_number = action_num

    def goal(self):
        goal = self.environment.goal()
        return self.processor.process_state(goal)

    def get_step(self, action, mode='q_learning', action_number=4):
        if mode == 'q_learning':
            degree_inc = 360.0 / action_number
            degree = action * degree_inc
            radian = math.radians(degree)
            step = [math.cos(radian), math.sin(radian)]
        else:
            norm = linalg.norm(action)
            if norm >= 1.0:
                action /= norm
            step = action
        return step

    def get_action(self, state, goal_state, policy):
        if self.her:
            s = np.concatenate((state, goal_state), axis=2)
        else:
            s = state

        if policy.use_network():
            action = self.model.predict(s)
        else:
            action = np.random.randint(0, self.action_number)

        return action

    def train(self, batch_size=32, max_action=200, max_episode=12000, warmup=120000):
        total_steps = 0
        self.callbacks.on_agent_begin(
            **{
                'agent_headers': ['episode_number', 'action_number', 'episode_reward'],
                'network_headers': ['loss']
            }
        )
        for episode_number in easy_range(1, max_episode):
            episode_reward = 0
            state = self.environment.reset()
            state = self.processor.process_state(state)
            self.callbacks.on_episode_begin(
                **{
                    'episode_number': episode_number,
                    'state': state
                }
            )

            goal_state = self.goal()

            for action_number in easy_range(1, max_action):
                action = self.get_action(state, goal_state, self.train_policy)
                self.callbacks.on_action_begin(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action
                    }
                )

                step = self.get_step(action, 'q_learning', self.action_number)
                next_state, reward, terminal, _ = self.environment.step(step)
                next_state = self.processor.process_state(next_state)
                if action_number >= max_action:
                    terminal = True

                self.callbacks.on_action_end(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action,
                        'reward': reward,
                        'terminal': terminal,
                        'next_state': next_state
                    }
                )

                processed_state = np.concatenate((state, goal_state), axis=2) if self.her else state
                clipped_reward = np.clip(reward - 0.25, -1, 1)
                processed_next_state = np.concatenate((next_state, goal_state), axis=2) if self.her else next_state
                self.memory.remember((processed_state, action, clipped_reward, processed_next_state, terminal))

                if total_steps > warmup:
                    self.train_policy.decay()
                    if total_steps % batch_size == 0:
                        self.callbacks.on_replay_begin()
                        mini_batch = self.memory.sample()
                        batch = self.processor.process_batch(mini_batch)
                        loss = self.model.train(batch)
                        self.callbacks.on_replay_end(
                            **{
                                'loss': loss
                            }
                        )

                episode_reward += reward
                state = copy.deepcopy(next_state)
                total_steps += 1

                if terminal:
                    self.callbacks.on_episode_end(
                        **{
                            'episode_number': episode_number,
                            'action_number': action_number,
                            'episode_reward': episode_reward
                        }
                    )
                    gc.collect()
                    break

        self.environment.close()
        self.callbacks.on_agent_end(
            **{
                'total_steps': total_steps
            }
        )

    def evaluate(self, max_action=50, max_episode=12):
        total_steps = 0
        self.callbacks.on_agent_begin()
        for episode_number in easy_range(1, max_episode):
            episode_reward = 0
            state = self.environment.reset()
            state = self.processor.process_state(state)
            self.callbacks.on_episode_begin(
                **{
                    'episode_number': episode_number,
                    'state': state
                }
            )

            goal_state = self.goal()

            for action_number in easy_range(1, max_action):
                action = self.get_action(state, goal_state, self.test_policy)
                self.callbacks.on_action_begin(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action
                    }
                )

                step = self.get_step(action, 'q_learning', self.action_number)
                next_state, reward, terminal, _ = self.environment.step(step)
                next_state = self.processor.process_state(next_state)
                if action_number >= max_action:
                    terminal = True

                self.callbacks.on_action_end(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action,
                        'reward': reward,
                        'terminal': terminal,
                        'next_state': next_state
                    }
                )

                episode_reward += reward
                state = copy.deepcopy(next_state)
                total_steps += 1

                if terminal:
                    self.callbacks.on_episode_end(
                        **{
                            'episode_number': episode_number,
                            'action_number': action_number,
                            'episode_reward': episode_reward
                        }
                    )
                    gc.collect()
                    break

        self.environment.close()
        self.callbacks.on_agent_end(
            **{
                'total_steps': total_steps
            }
        )


class DDPGTrainer:
    def __init__(self, environment, random_process, processor, memory, model, callbacks, her=False):
        super(DDPGTrainer, self).__init__()

        self.environment = environment
        self.memory = memory
        self.processor = processor
        self.model = model
        self.callbacks = callbacks
        self.random_process = random_process
        self.her = her

    def goal(self):
        goal = self.environment.goal()
        return self.processor.process_state(goal)

    def get_action(self, state, goal_state):
        if self.her:
            action = self.model.predict(np.concatenate((state, goal_state), axis=2))
        else:
            action = self.model.predict(state)
        action += self.random_process.sample()

        return action

    def train(self, batch_size=32, max_action=50, max_episode=120, warmup=0, replay_interval=4, update_interval=1, test_interval=1000):
        total_steps = 0
        self.callbacks.on_agent_begin(
            **{
                'agent_headers': ['episode_number', 'action_number', 'episode_reward'],
                'network_headers': ['actor_loss', 'critic_loss', 'critic_extra_loss']
            }
        )
        for episode_number in easy_range(1, max_episode):
            episode_reward = 0
            state = self.environment.reset()
            state = self.processor.process_state(state)
            self.callbacks.on_episode_begin(
                **{
                    'episode_number': episode_number,
                    'state': state
                }
            )

            goal_state = self.goal()

            for action_number in easy_range(1, max_action):
                action = self.get_action(state, goal_state)
                self.callbacks.on_action_begin(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action
                    }
                )

                if hasattr(self.environment, 'get_step'):
                    step = self.environment.get_step(action, 'policy_optimization')
                else:
                    step = action
                next_state, reward, terminal, _ = self.environment.step(step)
                next_state = self.processor.process_state(next_state)
                if action_number >= max_action:
                    terminal = True

                self.callbacks.on_action_end(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action,
                        'reward': reward,
                        'terminal': terminal,
                        'next_state': next_state
                    }
                )

                processed_state = np.concatenate((state, goal_state), axis=2) if self.her else state
                clipped_reward = np.clip(reward - 0.25, -1, 1)
                processed_next_state = np.concatenate((next_state, goal_state), axis=2) if self.her else next_state
                self.memory.remember(processed_state, action, clipped_reward, processed_next_state, terminal)

                if total_steps > warmup:
                    self.random_process.decay()
                    if total_steps % replay_interval == 0:
                        self.callbacks.on_replay_begin()
                        mini_batch = self.memory.sample()
                        batch = self.processor.process_batch(mini_batch)
                        loss = self.model.train(batch, ((total_steps - warmup) // replay_interval) % update_interval == 0)
                        self.callbacks.on_replay_end(
                            **{
                                'loss': loss
                            }
                        )

                episode_reward += reward
                state = deepcopy(next_state)
                total_steps += 1

                if terminal:
                    self.callbacks.on_episode_end(
                        **{
                            'episode_number': episode_number,
                            'action_number': action_number,
                            'episode_reward': episode_reward
                        }
                    )
                    gc.collect()
                    break

        self.environment.close()
        self.callbacks.on_agent_end(
            **{
                'total_steps': total_steps
            }
        )

    def evaluate(self, max_action=50, max_episode=12):
        total_steps = 0
        self.callbacks.on_agent_begin()
        for episode_number in easy_range(1, max_episode):
            episode_reward = 0
            state = self.environment.reset()
            state = self.processor.process_state(state)
            self.callbacks.on_episode_begin(
                **{
                    'episode_number': episode_number,
                    'state': state
                }
            )

            goal_state = self.goal()

            for action_number in easy_range(1, max_action):
                action = self.get_action(state, goal_state)
                self.callbacks.on_action_begin(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action
                    }
                )

                if hasattr(self.environment, 'get_step'):
                    step = self.environment.get_step(action, 'policy_optimization')
                else:
                    step = action
                next_state, reward, terminal, _ = self.environment.step(step)
                next_state = self.processor.process_state(next_state)
                if action_number >= max_action:
                    terminal = True

                self.callbacks.on_action_end(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action,
                        'reward': reward,
                        'terminal': terminal,
                        'next_state': next_state
                    }
                )

                episode_reward += reward
                state = deepcopy(next_state)
                total_steps += 1

                if terminal:
                    self.callbacks.on_episode_end(
                        **{
                            'episode_number': episode_number,
                            'action_number': action_number,
                            'episode_reward': episode_reward
                        }
                    )

                    gc.collect()
                    break

        self.environment.close()
        self.callbacks.on_agent_end(
            **{
                'total_steps': total_steps
            }
        )


class TD3Trainer:
    def __init__(self, environment, random_process, processor, memory, model, callbacks, her=False):
        super(TD3Trainer, self).__init__()

        self.environment = environment
        self.memory = memory
        self.processor = processor
        self.model = model
        self.callbacks = callbacks
        self.random_process = random_process
        self.her = her

    def goal(self):
        goal = self.environment.goal()
        return self.processor.process_state(goal)

    def get_action(self, state, goal_state):
        if self.her:
            action = self.model.predict(np.concatenate((state, goal_state), axis=2))
        else:
            action = self.model.predict(state)
        action += self.random_process.sample()

        return action

    def train(self, batch_size=32, max_action=50, max_episode=120, warmup=0, replay_interval=4, update_interval=1, test_interval=1000):
        total_steps = 0
        self.callbacks.on_agent_begin(
            **{
                'agent_headers': ['episode_number', 'action_number', 'episode_reward'],
                'network_headers': ['actor_loss', 'critic_loss', 'critic_extra_loss']
            }
        )
        for episode_number in easy_range(1, max_episode):
            episode_reward = 0
            state = self.environment.reset()
            state = self.processor.process_state(state)
            self.callbacks.on_episode_begin(
                **{
                    'episode_number': episode_number,
                    'state': state
                }
            )

            goal_state = self.goal()

            for action_number in easy_range(1, max_action):
                action = self.get_action(state, goal_state)
                self.callbacks.on_action_begin(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action
                    }
                )

                if hasattr(self.environment, 'get_step'):
                    step = self.environment.get_step(action, 'policy_optimization')
                else:
                    step = action
                next_state, reward, terminal, _ = self.environment.step(step)
                next_state = self.processor.process_state(next_state)
                if action_number >= max_action:
                    terminal = True

                self.callbacks.on_action_end(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action,
                        'reward': reward,
                        'terminal': terminal,
                        'next_state': next_state
                    }
                )

                processed_state = np.concatenate((state, goal_state), axis=2) if self.her else state
                clipped_reward = np.clip(reward - 0.25, -1, 1)
                processed_next_state = np.concatenate((next_state, goal_state), axis=2) if self.her else next_state
                self.memory.remember(processed_state, action, clipped_reward, processed_next_state, terminal)

                if total_steps > warmup:
                    self.random_process.decay()
                    if total_steps % replay_interval == 0:
                        self.callbacks.on_replay_begin()
                        mini_batch = self.memory.sample()
                        batch = self.processor.process_batch(mini_batch)
                        loss = self.model.train(batch, ((total_steps - warmup) // replay_interval) % update_interval == 0)
                        self.callbacks.on_replay_end(
                            **{
                                'loss': loss
                            }
                        )

                episode_reward += reward
                state = deepcopy(next_state)
                total_steps += 1

                if terminal:
                    self.callbacks.on_episode_end(
                        **{
                            'episode_number': episode_number,
                            'action_number': action_number,
                            'episode_reward': episode_reward
                        }
                    )
                    gc.collect()
                    break

        self.environment.close()
        self.callbacks.on_agent_end(
            **{
                'total_steps': total_steps
            }
        )

    def evaluate(self, max_action=50, max_episode=12):
        total_steps = 0
        self.callbacks.on_agent_begin()
        for episode_number in easy_range(1, max_episode):
            episode_reward = 0
            state = self.environment.reset()
            state = self.processor.process_state(state)
            self.callbacks.on_episode_begin(
                **{
                    'episode_number': episode_number,
                    'state': state
                }
            )

            goal_state = self.goal()

            for action_number in easy_range(1, max_action):
                action = self.get_action(state, goal_state)
                self.callbacks.on_action_begin(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action
                    }
                )

                if hasattr(self.environment, 'get_step'):
                    step = self.environment.get_step(action, 'policy_optimization')
                else:
                    step = action
                next_state, reward, terminal, _ = self.environment.step(step)
                next_state = self.processor.process_state(next_state)
                if action_number >= max_action:
                    terminal = True

                self.callbacks.on_action_end(
                    **{
                        'episode_number': episode_number,
                        'action_number': action_number,
                        'state': state,
                        'action': action,
                        'reward': reward,
                        'terminal': terminal,
                        'next_state': next_state
                    }
                )

                episode_reward += reward
                state = deepcopy(next_state)
                total_steps += 1

                if terminal:
                    self.callbacks.on_episode_end(
                        **{
                            'episode_number': episode_number,
                            'action_number': action_number,
                            'episode_reward': episode_reward
                        }
                    )

                    gc.collect()
                    break

        self.environment.close()
        self.callbacks.on_agent_end(
            **{
                'total_steps': total_steps
            }
        )



import ctypes
import struct
import sys
from ctypes import wintypes

if sys.platform != 'win32':
    raise Exception('The ram module should only be used on a Windows system.')


def type_unpack(of_type):
    of_type = of_type.lower()
    if of_type == 'short':
        structure_type = 'h'
        structure_length = 2
    elif of_type == 'ushort':
        structure_type = 'H'
        structure_length = 2
    elif of_type == 'int':
        structure_type = 'i'
        structure_length = 4
    elif of_type == 'uint':
        structure_type = 'I'
        structure_length = 4
    elif of_type == 'long':
        structure_type = 'l'
        structure_length = 4
    elif of_type == 'ulong':
        structure_type = 'L'
        structure_length = 4
    elif of_type == 'float':
        structure_type = 'f'
        structure_length = 4
    elif of_type == 'double':
        structure_type = 'd'
        structure_length = 8
    else:
        raise TypeError(f'Unknown type {of_type}')
    return '<' + structure_type, structure_length


def hex_dump(data, address=0, prefix='', of_type='bytes'):
    dump = prefix
    piece = ''
    if of_type != 'bytes':
        structure_type, structure_len = type_unpack(of_type)
        for i in range(0, len(data), structure_len):
            if address % 16 == 0:
                dump += ' '
                for char in piece:
                    if 32 <= ord(char) <= 126:
                        dump += char
                    else:
                        dump += '.'

                dump += '\n%s%08X: ' % (prefix, address)
                piece = ''
            temp_val = 'NaN'
            try:
                packed_val = data[i:i + structure_len]
                temp_val = struct.unpack(structure_type, packed_val)[0]
            except Exception as ex_hex_dump:
                print(str(ex_hex_dump))

            if temp_val == 'NaN':
                dump += '{:<15} '.format(temp_val)
            elif of_type == 'float':
                dump += '{:<15.4f} '.format(temp_val)
            else:
                dump += '{:<15} '.format(temp_val)
            address += structure_len
    else:
        for byte in data:
            if address % 16 == 0:
                dump += ' '
                for char in piece:
                    if 32 <= ord(char) <= 126:
                        dump += char
                    else:
                        dump += '.'

                dump += '\n%s%08X: ' % (prefix, address)
                piece = ''

            dump += '%02X ' % byte
            piece += chr(byte)
            address += 1

    remainder = address % 16
    if remainder != 0:
        dump += '   ' * (16 - remainder) + ' '
    for char in piece:
        if 32 <= ord(char) <= 126:
            dump += char
        else:
            dump += '.'

    return dump + '\n'


MEM_COMMIT = 4096
MEM_FREE = 65536
MEM_RESERVE = 8192

PAGE_EXECUTE_READWRITE = 64
PAGE_EXECUTE_READ = 32
PAGE_READONLY = 2
PAGE_READWRITE = 4
PAGE_NOCACHE = 512
PAGE_WRITECOMBINE = 1024
PAGE_GUARD = 256

UNPROTECTED_DACL_SECURITY_INFORMATION = 536870912
DACL_SECURITY_INFORMATION = 4

if ctypes.sizeof(ctypes.c_void_p) == 8:
    ULONG_PTR = ctypes.c_ulonglong
else:
    ULONG_PTR = ctypes.c_ulong


class ModuleEntry32(ctypes.Structure):
    _fields_ = [
        ('dwSize', ctypes.c_ulong),
        ('th32ModuleID', ctypes.c_ulong),
        ('th32ProcessID', ctypes.c_ulong),
        ('GlblcntUsage', ctypes.c_ulong),
        ('ProccntUsage', ctypes.c_ulong),
        ('modBaseAddr', ctypes.POINTER(ctypes.c_byte)),
        ('modBaseSize', ctypes.c_ulong),
        ('hModule', ctypes.c_void_p),
        ('szModule', ctypes.c_char * 256),
        ('szExePath', ctypes.c_char * 260)
    ]


class MemoryBasicInformation(ctypes.Structure):
    _fields_ = [
        ('BaseAddress', ctypes.c_void_p),
        ('AllocationBase', ctypes.c_void_p),
        ('AllocationProtect', wintypes.DWORD),
        ('RegionSize', ctypes.c_size_t),
        ('State', wintypes.DWORD),
        ('Protect', wintypes.DWORD),
        ('Type', wintypes.DWORD)
    ]


class SystemInfo(ctypes.Structure):
    _fields_ = [
        ('wProcessorArchitecture', wintypes.WORD),
        ('wReserved', wintypes.WORD),
        ('dwPageSize', wintypes.DWORD),
        ('lpMinimumApplicationAddress', wintypes.LPVOID),
        ('lpMaximumApplicationAddress', wintypes.LPVOID),
        ('dwActiveProcessorMask', ULONG_PTR),
        ('dwNumberOfProcessors', wintypes.DWORD),
        ('dwProcessorType', wintypes.DWORD),
        ('dwAllocationGranularity', wintypes.DWORD),
        ('wProcessorLevel', wintypes.WORD),
        ('wProcessorRevision', wintypes.WORD)
    ]


class SecurityDescriptor(ctypes.Structure):
    _fields_ = [
        ('SID', wintypes.DWORD),
        ('group', wintypes.DWORD),
        ('dacl', wintypes.DWORD),
        ('sacl', wintypes.DWORD),
        ('test', wintypes.DWORD)
    ]


class Th32csClass(object):
    INHERIT = 2147483648
    SNAPHEAPLIST = 1
    SNAPMODULE = 8
    SNAPMODULE32 = 16
    SNAPPROCESS = 2
    SNAPTHREAD = 4
    ALL = 2032639


PSecurityDescriptor = ctypes.POINTER(SecurityDescriptor)

CreateToolhelp32Snapshot = ctypes.windll.kernel32.CreateToolhelp32Snapshot
CreateToolhelp32Snapshot.argtypes = [
    ctypes.c_int, ctypes.c_int
]
CreateToolhelp32Snapshot.reltype = ctypes.c_long

Module32First = ctypes.windll.kernel32.Module32First
Module32First.argtypes = [
    ctypes.c_void_p, ctypes.POINTER(ModuleEntry32)
]
Module32First.rettype = ctypes.c_int

Module32Next = ctypes.windll.kernel32.Module32Next
Module32Next.argtypes = [
    ctypes.c_void_p, ctypes.POINTER(ModuleEntry32)
]
Module32Next.rettype = ctypes.c_int

ReadProcessMemory = ctypes.windll.kernel32.ReadProcessMemory
ReadProcessMemory.argtypes = [
    wintypes.HANDLE, wintypes.LPCVOID, wintypes.LPVOID, ctypes.c_size_t, ctypes.POINTER(ctypes.c_size_t)
]

VirtualQueryEx = ctypes.windll.kernel32.VirtualQueryEx
VirtualQueryEx.argtypes = [
    wintypes.HANDLE, wintypes.LPCVOID, ctypes.POINTER(MemoryBasicInformation), ctypes.c_size_t
]
VirtualQueryEx.restype = ctypes.c_size_t

IsWow64Process = None
if hasattr(ctypes.windll.kernel32, 'IsWow64Process'):
    IsWow64Process = ctypes.windll.kernel32.IsWow64Process
    IsWow64Process.argtypes = [
        ctypes.c_void_p, ctypes.POINTER(ctypes.c_bool)
    ]
    IsWow64Process.restype = ctypes.c_bool

if ctypes.sizeof(ctypes.c_void_p) == 8:
    NtWow64ReadVirtualMemory64 = None
else:
    try:
        NtWow64ReadVirtualMemory64 = ctypes.windll.ntdll.NtWow64ReadVirtualMemory64
        NtWow64ReadVirtualMemory64.argtypes = [
            wintypes.HANDLE, ctypes.c_longlong, wintypes.LPVOID, ctypes.c_ulonglong, ctypes.POINTER(ctypes.c_ulong)
        ]
        NtWow64ReadVirtualMemory64.restype = wintypes.BOOL
    except Exception as ex:
        print(str(ex))
        NtWow64ReadVirtualMemory64 = None



from pykafka import KafkaClient
from pykafka.common import OffsetType

from joatmon.plugin.message.core import MessagePlugin


class KafkaPlugin(MessagePlugin):
    def __init__(self, host):
        self.client = KafkaClient(host)

    def get_producer(self, topic):
        return self.client.topics[topic].get_producer()

    def get_consumer(self, topic):
        return self.client.topics[topic].get_simple_consumer(
            auto_commit_enable=True,
            consumer_timeout_ms=1000,
            consumer_group='my_group',
            auto_commit_interval_ms=1000,
            auto_offset_reset=OffsetType.LATEST
        )



import datetime
import json
import os

from joatmon.plugin.logger.core import LoggerPlugin


class FileLogger(LoggerPlugin):
    def __init__(self, level: str, base_folder: str, language, ip):
        super(FileLogger, self).__init__(level, language, ip)

        self.folder = base_folder

    async def _write(self, log: dict):
        dt = datetime.datetime.now()
        folder = os.path.join(self.folder, dt.strftime(f'%Y{os.sep}%m{os.sep}'))
        if not os.path.exists(folder):
            os.makedirs(folder)
        file = os.path.join(folder, dt.strftime(f'%d') + '.log')
        with open(file, 'a') as f:
            f.write(json.dumps(log) + os.linesep)



from collections import deque


def dfs(graph, v, discovered):
    stack = deque()
    stack.append(v)

    while stack:
        v = stack.pop()
        if discovered[v]:
            continue

        discovered[v] = True
        yield v

        # do for every edge (v, u)
        adj_list = graph.adj_list[v]
        for i in reversed(range(len(adj_list))):
            u = adj_list[i]
            if not discovered[u]:
                stack.append(u)


def dfs_r(graph, v, discovered):
    discovered[v] = True  # mark the current node as discovered
    yield v

    for u in graph.adj_list[v]:
        if not discovered[u]:  # if `u` is not yet discovered
            dfs_r(graph, u, discovered)



import copy
import struct

from joatmon.system.memory.address import Address


class Locator(object):
    def __init__(self, worker, of_type='unknown', start=None, end=None):
        self.worker = worker
        self.of_type = of_type
        self.last_iteration = {}
        self.last_value = None
        self.start = start
        self.end = end

    def find(self, value, erase_last=True):
        return self.feed(value, erase_last)

    def feed(self, value, erase_last=True):
        self.last_value = value
        new_iter = copy.copy(self.last_iteration)
        if self.of_type == 'unknown':
            all_types = ['uint', 'int', 'long', 'ulong', 'float', 'double', 'short', 'ushort']
        else:
            all_types = [self.of_type]

        for of_type in all_types:
            if of_type not in new_iter:
                try:
                    new_iter[of_type] = [
                        Address(address, self.worker.process, of_type) for address in
                        self.worker.memory_search(value, of_type, start_offset=self.start, end_offset=self.end)
                    ]
                except struct.error:
                    new_iter[of_type] = []
            else:
                addresses = []
                for address in new_iter[of_type]:
                    try:
                        found = self.worker.process.read(address, of_type)
                        if int(found) == int(value):
                            addresses.append(Address(address, self.worker.process, of_type))
                    except Exception as ex:
                        print(str(ex))

                new_iter[of_type] = addresses

        if erase_last:
            del self.last_iteration
            self.last_iteration = new_iter
        return new_iter

    def get_addresses(self):
        return self.last_iteration

    def diff(self, erase_last=False):
        return self.get_modified_address(erase_last)

    def get_modified_address(self, erase_last=False):
        last = self.last_iteration
        new = self.feed(self.last_value, erase_last=erase_last)
        ret = {}
        for of_type, addresses in last.items():
            typeset = set(new[of_type])
            for _address in addresses:
                if _address not in typeset:
                    if of_type not in ret:
                        ret[of_type] = []
                    ret[of_type].append(_address)

        return ret



from joatmon.serializable import Serializable


class ValidationException(Exception):
    pass


class Constraint(Serializable):
    def __init__(self, field, validator=None):
        super(Constraint, self).__init__()

        self.field = field
        self.validator = validator

    @staticmethod
    def create(constraint_type, **kwargs):
        if constraint_type == 'length':
            return LengthConstraint(**kwargs)
        if constraint_type == 'integer':
            return IntegerValueConstraint(**kwargs)
        if constraint_type == 'unique':
            return UniqueConstraint(**kwargs)
        if constraint_type == 'custom':
            return CustomConstraint(**kwargs)

    def check(self, obj):
        if self.validator is not None:
            if callable(self.validator):
                if self.validator(obj):
                    return True
                else:
                    raise ValidationException(f'field.{self.field}.not_valid')
            else:
                return True
        else:
            return True


class LengthConstraint(Constraint):
    def __init__(self, field, min_length=None, max_length=None):
        if min_length is None and max_length is None:
            raise ValueError('at least one of the min_length, max_length has to be provided')

        if min_length is None:
            validator = lambda x: len(x) <= max_length
        elif max_length is None:
            validator = lambda x: min_length <= len(x)
        else:
            validator = lambda x: min_length <= len(x) <= max_length

        super(LengthConstraint, self).__init__(field, validator)

        self.min_length = min_length
        self.max_length = max_length


class IntegerValueConstraint(Constraint):
    def __init__(self, field, min_value=None, max_value=None):
        if min_value is None and max_value is None:
            raise ValueError('at least one of the min_value, max_value has to be provided')

        if min_value is None:
            validator = lambda x: x <= max_value
        elif max_value is None:
            validator = lambda x: min_value <= x
        else:
            validator = lambda x: min_value <= x <= max_value

        super(IntegerValueConstraint, self).__init__(field, validator)

        self.min_value = min_value
        self.max_value = max_value


class PrimaryKeyConstraint(Constraint):
    def __init__(self, field):
        super(PrimaryKeyConstraint, self).__init__(field)


class ForeignKeyConstrain(Constraint):
    def __enter__(self, field):
        super(ForeignKeyConstrain, self).__init__(field)


# unique constraint should have more than one field
class UniqueConstraint(Constraint):
    def __init__(self, field):
        super(UniqueConstraint, self).__init__(field)


class CustomConstraint(Constraint):
    def __init__(self, field, validator=lambda x: True):
        super(CustomConstraint, self).__init__(field, validator)



from joatmon import context


def register(cls, alias, *args, **kwargs):
    context.set_value(alias.replace('-', '_'), PluginProxy(cls, *args, **kwargs))


class Plugin:
    ...


class PluginProxy:
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

        self.instance = None

    def __getattr__(self, item):
        if self.instance is None:
            self.instance = self.cls(*self.args, **self.kwargs)
        return getattr(self.instance, item.replace('-', '_'))

    async def __aenter__(self):
        if self.instance is None:
            self.instance = self.cls(*self.args, **self.kwargs)
        return await self.instance.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.instance is None:
            self.instance = self.cls(*self.args, **self.kwargs)
        return await self.instance.__aexit__(exc_type, exc_val, exc_tb)



from ..errors import HZMQTTException
from ..mqtt.packet import (
    CONNACK,
    ConnackPacket,
    CONNECT,
    ConnectPacket,
    DISCONNECT,
    DisconnectPacket,
    MQTTFixedHeader,
    PINGREQ,
    PingReqPacket,
    PINGRESP,
    PingRespPacket,
    PUBACK,
    PubackPacket,
    PUBCOMP,
    PubcompPacket,
    PUBLISH,
    PublishPacket,
    PUBREC,
    PubrecPacket,
    PUBREL,
    PubrelPacket,
    SUBACK,
    SubackPacket,
    SUBSCRIBE,
    SubscribePacket,
    UNSUBACK,
    UnsubackPacket,
    UNSUBSCRIBE,
    UnsubscribePacket
)

packet_dict = {
    CONNECT: ConnectPacket,
    CONNACK: ConnackPacket,
    PUBLISH: PublishPacket,
    PUBACK: PubackPacket,
    PUBREC: PubrecPacket,
    PUBREL: PubrelPacket,
    PUBCOMP: PubcompPacket,
    SUBSCRIBE: SubscribePacket,
    SUBACK: SubackPacket,
    UNSUBSCRIBE: UnsubscribePacket,
    UNSUBACK: UnsubackPacket,
    PINGREQ: PingReqPacket,
    PINGRESP: PingRespPacket,
    DISCONNECT: DisconnectPacket
}


def packet_class(fixed_header: MQTTFixedHeader):
    try:
        cls = packet_dict[fixed_header.packet_type]
        return cls
    except KeyError:
        raise HZMQTTException("Unexpected packet Type '%s'" % fixed_header.packet_type)



class CoreNetwork(object):
    """
    Abstract base class for all implemented nn.

    Do not use this abstract base class directly
    but instead use one of the concrete nn implemented.

    To implement your own nn, you have to implement the following methods:

    - `act`
    - `replay`
    - `load`
    - `save`
    """

    def __init__(self):
        super(CoreNetwork, self).__init__()

    def load(self):
        """
        load
        """
        raise NotImplementedError

    def save(self):
        """
        save
        """
        raise NotImplementedError

    def predict(self):
        """
        Get the action for given state.

        Accepts a state and returns an abstract action.

        # Arguments
            state (abstract): Current state of the game.

        # Returns
            action (abstract): Network's predicted action for given state.
        """
        raise NotImplementedError

    def train(self):
        """
        Train the nn with given batch.

        # Arguments
            batch (abstract): Mini Batch from Experience Replay Memory.
        """
        raise NotImplementedError

    def evaluate(self):
        """
        evaluate
        """
        raise NotImplementedError



import uuid

from couchbase.auth import PasswordAuthenticator
from couchbase.cluster import Cluster
from couchbase.logic.n1ql import QueryScanConsistency
from couchbase.options import (
    ClusterOptions,
    QueryOptions
)

from joatmon.plugin.database.core import DatabasePlugin


class CouchBaseDatabase(DatabasePlugin):
    DATABASES = set()
    CREATED_COLLECTIONS = set()
    UPDATED_COLLECTIONS = set()

    def __init__(self, uri, bucket, scope, username, password):
        auth = PasswordAuthenticator(username, password)

        self.bucket = bucket
        self.scope = scope

        self.db = Cluster(uri, ClusterOptions(auth)).bucket(bucket).scope(scope)

    async def _create_collection(self, collection):
        self.db.query(
            f'CREATE COLLECTION `{self.bucket}`.{self.scope}.{collection.__collection__} '
            f'if not exists',
            QueryOptions(scan_consistency=QueryScanConsistency.REQUEST_PLUS)
        ).execute()

    async def insert(self, document, *docs):
        for doc in docs:
            await self._create_collection(document.__metaclass__)

            values = '{' + ', '.join([f'"{k}" : ${k}' for k, v in doc.items()]) + '}'
            self.db.query(
                f'insert into `{self.bucket}`.{self.scope}.{document.__metaclass__.__collection__} '
                f'(key, value) '
                f'values '
                f'(\'{str(uuid.uuid4())}\', {values}) '
                f'returning *',
                QueryOptions(named_parameters={k: str(v) if isinstance(v, uuid.UUID) else v for k, v in doc.items()}, scan_consistency=QueryScanConsistency.REQUEST_PLUS)
            ).execute()

    async def read(self, document, query):
        await self._create_collection(document.__metaclass__)

        values = 'and '.join([f'`{k}` = ${k}' for k, v in query.items()])
        if len(values) > 0:
            result = self.db.query(
                f'SELECT * '
                f'FROM `{self.bucket}`.{self.scope}.{document.__metaclass__.__collection__} '
                f'WHERE {values}',
                QueryOptions(named_parameters={k: str(v) if isinstance(v, uuid.UUID) else v for k, v in query.items()}, scan_consistency=QueryScanConsistency.REQUEST_PLUS)
            ).execute()
        else:
            result = self.db.query(
                f'SELECT * '
                f'FROM `{self.bucket}`.{self.scope}.{document.__metaclass__.__collection__} ',
                QueryOptions(named_parameters={k: str(v) if isinstance(v, uuid.UUID) else v for k, v in query.items()}, scan_consistency=QueryScanConsistency.REQUEST_PLUS)
            ).execute()
        for r in result:
            yield document(**r[document.__metaclass__.__collection__])

    async def update(self, document, query, update):
        await self._create_collection(document.__metaclass__)

        query_values = 'and '.join([f'`{k}` = $query_{k}' for k, v in query.items()])
        update_values = ', '.join([f'`{k}` = $update_{k}' for k, v in update.items()])

        params = {}
        params.update({f'query_{k}': str(v) if isinstance(v, uuid.UUID) else v for k, v in query.items()})
        params.update({f'update_{k}': str(v) if isinstance(v, uuid.UUID) else v for k, v in update.items()})
        self.db.query(
            f'update `{self.bucket}`.{self.scope}.{document.__metaclass__.__collection__} '
            f'set {update_values} '
            f'WHERE {query_values}',
            QueryOptions(named_parameters=params, scan_consistency=QueryScanConsistency.REQUEST_PLUS)
        ).execute()

    async def delete(self, document, query):
        await self._create_collection(document.__metaclass__)

        values = 'and '.join([f'`{k}` = ${k}' for k, v in query.items()])
        self.db.query(
            f'delete from `{self.bucket}`.{self.scope}.{document.__metaclass__.__collection__} '
            f'WHERE {values}',
            QueryOptions(named_parameters={k: str(v) if isinstance(v, uuid.UUID) else v for k, v in query.items()}, scan_consistency=QueryScanConsistency.REQUEST_PLUS)
        ).execute()

    async def start(self):
        ...

    async def commit(self):
        ...

    async def abort(self):
        ...

    async def end(self):
        ...



from joatmon.system.memory.core import hex_dump


class Address(object):
    def __init__(self, value, process, default_type='uint'):
        self.value = int(value)
        self.process = process
        self.default_type = default_type
        self.symbolic_name = None

    def read(self, of_type=None, max_length=None, errors='raise'):
        if max_length is None:
            try:
                max_length = int(of_type)
                of_type = None
            except Exception as ex:
                print(str(ex))

        if not of_type:
            of_type = self.default_type

        if not max_length:
            return self.process.read(self.value, of_type=of_type, errors=errors)
        else:
            return self.process.read(self.value, of_type=of_type, max_length=max_length, errors=errors)

    def write(self, data, of_type=None):
        if not of_type:
            of_type = self.default_type
        return self.process.write(self.value, data, of_type=of_type)

    def symbol(self):
        return self.process.get_symbolic_name(self.value)

    def get_instruction(self):
        return self.process.get_instruction(self.value)

    def dump(self, of_type='bytes', size=512, before=32):
        buf = self.process.read_bytes(self.value - before, size)
        print(hex_dump(buf, self.value - before, of_type=of_type))

    def __nonzero__(self):
        return self.value is not None and self.value != 0

    def __add__(self, other):
        return Address(self.value + int(other), self.process, self.default_type)

    def __sub__(self, other):
        return Address(self.value - int(other), self.process, self.default_type)

    def __repr__(self):
        if not self.symbolic_name:
            self.symbolic_name = self.symbol()
        return str(f'<Addr: {self.symbolic_name}>')

    def __str__(self):
        if not self.symbolic_name:
            self.symbolic_name = self.symbol()
        return str(f'<Addr: {self.symbolic_name} : "{str(self.read())}" ({self.default_type})>')

    def __int__(self):
        return int(self.value)

    def __hex__(self):
        return hex(self.value)

    def __get__(self, instance, owner):
        return self.value

    def __set__(self, instance, value):
        self.value = int(value)

    def __lt__(self, other):
        return self.value < int(other)

    def __le__(self, other):
        return self.value <= int(other)

    def __eq__(self, other):
        return self.value == int(other)

    def __ne__(self, other):
        return self.value != int(other)

    def __gt__(self, other):
        return self.value > int(other)

    def __ge__(self, other):
        return self.value >= int(other)



import dataclasses
import math

import pygame
from chess import (
    BB_LIGHT_SQUARES,
    BB_SQUARES,
    BISHOP,
    Board,
    FILE_NAMES,
    KNIGHT,
    Move,
    PAWN,
    QUEEN,
    RANK_NAMES,
    ROOK,
    Square,
    WHITE
)
from pygame.color import THECOLORS

__all__ = ['ChessEnv', 'create_move_labels']


@dataclasses.dataclass(unsafe_hash=True)
class Arrow:
    from_square: Square
    to_square: Square
    color: str = None

    def __eq__(self, other):
        return self.from_square == other.from_square and self.to_square == other.to_square and self.color == other.color


def arrow(screen, color, start, end, thickness):
    # arrow body
    pygame.draw.line(
        screen,
        color,
        start,
        end,
        thickness
    )
    rotation = (math.atan2(start[1] - end[1], end[0] - start[0])) + math.pi / 2
    # arrow head
    pygame.draw.polygon(
        screen,
        color,
        (
            (end[0] + 9 * math.sin(rotation), end[1] + 9 * math.cos(rotation)),
            (end[0] + 9 * math.sin(rotation - 120 * math.pi / 180), end[1] + 9 * math.cos(rotation - 120 * math.pi / 180)),
            (end[0] + 9 * math.sin(rotation + 120 * math.pi / 180), end[1] + 9 * math.cos(rotation + 120 * math.pi / 180))
        )
    )


setattr(pygame.draw, 'arrow', arrow)

pygame.init()

SQUARE_SIZE = 64
MARGIN = 20
empty_space = 40
light_border_size = 5
dark_border_size = 3
piece_size = 64

PIECES = {
    "b": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/black-bishop.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "k": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/black-king.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "n": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/black-knight.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "p": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/black-pawn.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "q": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/black-queen.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "r": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/black-rook.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "B": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/white-bishop.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "K": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/white-king.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "N": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/white-knight.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "P": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/white-pawn.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "Q": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/white-queen.png'), (SQUARE_SIZE, SQUARE_SIZE)),
    "R": pygame.transform.scale(pygame.image.load('joatmon/game/assets/chess/white-rook.png'), (SQUARE_SIZE, SQUARE_SIZE))
}


def create_move_labels():
    labels_array = []
    letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
    numbers = ['1', '2', '3', '4', '5', '6', '7', '8']
    promoted_to = ['q', 'r', 'b', 'n']

    for l1 in range(8):
        for n1 in range(8):
            destinations = [(t, n1) for t in range(8)] + \
                           [(l1, t) for t in range(8)] + \
                           [(l1 + t, n1 + t) for t in range(-7, 8)] + \
                           [(l1 + t, n1 - t) for t in range(-7, 8)] + \
                           [(l1 + a, n1 + b) for (a, b) in [(-2, -1), (-1, -2), (-2, 1), (1, -2), (2, -1), (-1, 2), (2, 1), (1, 2)]]
            for (l2, n2) in destinations:
                if (l1, n1) != (l2, n2) and l2 in range(8) and n2 in range(8):
                    move = letters[l1] + numbers[n1] + letters[l2] + numbers[n2]
                    labels_array.append(move)
    for l1 in range(8):
        letter = letters[l1]
        for p in promoted_to:
            labels_array.append(letter + '2' + letter + '1' + p)
            labels_array.append(letter + '7' + letter + '8' + p)
            if l1 > 0:
                l_l = letters[l1 - 1]
                labels_array.append(letter + '2' + l_l + '1' + p)
                labels_array.append(letter + '7' + l_l + '8' + p)
            if l1 < 7:
                l_r = letters[l1 + 1]
                labels_array.append(letter + '2' + l_r + '1' + p)
                labels_array.append(letter + '7' + l_r + '8' + p)
    return labels_array


class ChessEnv:
    def __init__(self):
        self.board = Board()
        self.board_size = empty_space + light_border_size + dark_border_size + piece_size * 8 + dark_border_size + light_border_size + empty_space
        self.screen = pygame.display.set_mode((self.board_size, self.board_size))

        self.reset()

    def __eq__(self, env):
        if isinstance(env, ChessEnv):
            return self.board == env.board
        else:
            return NotImplemented

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        env = self.copy()
        memo[id(self)] = env
        return env

    def copy(self, *, stack=True):
        env = type(self)()
        env.board = env.board.copy(stack=stack)
        return env

    def reset(self):
        self.board.reset()
        return self.get_observation()

    def get_observation(self, orientation=WHITE, flipped=False, mode='str'):
        if mode == 'rgb':
            orientation ^= flipped

            surface = pygame.Surface((piece_size * 8, piece_size * 8))

            surface.fill(THECOLORS['black'])

            for square, bb in enumerate(BB_SQUARES):
                file_index = square & 7
                rank_index = square >> 3

                x = (file_index if orientation else 7 - file_index) * piece_size
                y = (7 - rank_index if orientation else rank_index) * piece_size

                color = "white" if BB_LIGHT_SQUARES & bb else "grey"

                pygame.draw.polygon(
                    surface,
                    THECOLORS[color],
                    [
                        (x, y),
                        (x, y + piece_size),
                        (x + piece_size, y + piece_size),
                        (x + piece_size, y)
                    ]
                )

                _piece = self.board.piece_at(square)

                if _piece is not None:
                    surface.blit(
                        PIECES[_piece.symbol()],
                        (
                            x,
                            y
                        )
                    )

            return surface
        else:
            state = []
            for square, bb in enumerate(BB_SQUARES):
                _piece = self.board.piece_at(square)
                if _piece is not None:
                    state.append(_piece.piece_type * (1 if _piece.color else -1))
                else:
                    state.append(0)
            return state

    def get_reward(self, piece):
        reward = 0
        if piece == PAWN:
            reward += 1
        if piece == KNIGHT:
            reward += 3.05
        if piece == BISHOP:
            reward += 3.33
        if piece == ROOK:
            reward += 5.63
        if piece == QUEEN:
            reward += 9.5
        if self.board.is_checkmate():
            reward += 100
        if self.board.turn:
            reward *= -1
        return reward

    def step(self, move, move_type='uci'):
        if move_type == 'san':
            move = self.board.parse_san(self.board.san(move))
        captured_piece_type = self.board.piece_at(move.to_square)
        return self.get_observation(), self.get_reward(captured_piece_type), self.board.is_game_over(), {'turn': not self.board.turn}

    def result(self, *, claim_draw=False):
        if self.board.is_checkmate():
            return "0-1" if self.board.turn == WHITE else "1-0"

        if claim_draw and self.board.can_claim_draw():
            return "1/2-1/2"

        if self.board.is_seventyfive_moves() or self.board.is_fivefold_repetition():
            return "1/2-1/2"

        if self.board.is_insufficient_material():
            return "1/2-1/2"

        if not any(self.board.generate_legal_moves()):
            return "1/2-1/2"

        return "*"

    def unicode(self, *, invert_color=False, borders=False, empty_square="⭘"):
        builder = []
        for rank_index in range(7, -1, -1):
            if borders:
                builder.append("  ")
                builder.append("-" * 17)
                builder.append("\n")

                builder.append(RANK_NAMES[rank_index])
                builder.append(" ")

            for file_index in range(8):
                square_index = rank_index * 8 + file_index

                if borders:
                    builder.append("|")
                elif file_index > 0:
                    builder.append(" ")

                _piece = self.board.piece_at(square_index)

                if _piece:
                    builder.append(_piece.unicode_symbol(invert_color=invert_color))
                else:
                    builder.append(empty_square)

            if borders:
                builder.append("|")

            if borders or rank_index > 0:
                builder.append("\n")

        if borders:
            builder.append("  ")
            builder.append("-" * 17)
            builder.append("\n")
            builder.append("   a b c d e f g h")

        return "".join(builder)

    def render(self, *, orientation=WHITE, flipped=False, mode='str'):
        if mode == 'rgb':
            orientation ^= flipped

            self.screen.fill(THECOLORS['grey'])

            font = pygame.font.SysFont(None, 36)

            for idx, file in enumerate(FILE_NAMES):
                img = font.render(file, True, THECOLORS['white'])
                self.screen.blit(
                    img,
                    (
                        empty_space + light_border_size + dark_border_size + piece_size // 2 + idx * piece_size,
                        empty_space + light_border_size + dark_border_size + piece_size // 4 + 8 * piece_size
                    )
                )

            for idx, rank in enumerate(RANK_NAMES):
                img = font.render(rank, True, THECOLORS['white'])
                self.screen.blit(
                    img,
                    (
                        empty_space // 2,
                        self.board_size + piece_size // 4 - empty_space - light_border_size - dark_border_size - (idx + 1) * piece_size
                    )
                )

            pygame.draw.polygon(
                self.screen,
                THECOLORS['white'],
                [
                    (empty_space, empty_space),
                    (empty_space, -empty_space + self.board_size),
                    (-empty_space + self.board_size, -empty_space + self.board_size),
                    (-empty_space + self.board_size, empty_space)
                ]
            )

            pygame.draw.polygon(
                self.screen,
                THECOLORS['black'],
                [
                    (empty_space + light_border_size, empty_space + light_border_size),
                    (empty_space + light_border_size, -empty_space - light_border_size + self.board_size),
                    (-empty_space - light_border_size + self.board_size, -empty_space - light_border_size + self.board_size),
                    (-empty_space - light_border_size + self.board_size, empty_space + light_border_size)
                ]
            )

            last_move = self.board.peek() or Move(from_square=-1, to_square=-1)
            arrows = set()
            arrows.add(Arrow(from_square=last_move.from_square, to_square=last_move.to_square, color='green'))

            for square, bb in enumerate(BB_SQUARES):
                file_index = square & 7
                rank_index = square >> 3

                x = (file_index if orientation else 7 - file_index) * piece_size + empty_space + light_border_size + dark_border_size
                y = (7 - rank_index if orientation else rank_index) * piece_size + empty_space + light_border_size + dark_border_size

                color = "white" if BB_LIGHT_SQUARES & bb else "grey"

                if self.board.piece_at(square) is not None and self.board.piece_at(square).color == self.board.turn:
                    for attacker in self.board.attackers(not self.board.turn, square):
                        if self.board.piece_at(attacker) is not None and self.board.piece_at(attacker).color != self.board.turn:
                            arrows.add(Arrow(from_square=attacker, to_square=square, color='red'))

                if self.board.piece_at(square) is not None and self.board.piece_at(square).color == self.board.turn:
                    for attack in self.board.attacks(square):
                        if self.board.piece_at(attack) is not None and self.board.piece_at(attack).color != self.board.turn:
                            arrows.add(Arrow(from_square=square, to_square=attack, color='orange'))

                pygame.draw.polygon(
                    self.screen,
                    THECOLORS[color],
                    [
                        (x, y),
                        (x, y + piece_size),
                        (x + piece_size, y + piece_size),
                        (x + piece_size, y)
                    ]
                )

                _piece = self.board.piece_at(square)

                if _piece is not None:
                    self.screen.blit(
                        PIECES[_piece.symbol()],
                        (
                            x,
                            y
                        )
                    )

            for _arrow in arrows:
                from_file_index = _arrow.from_square & 7
                from_rank_index = _arrow.from_square >> 3

                to_file_index = _arrow.to_square & 7
                to_rank_index = _arrow.to_square >> 3

                from_x = (from_file_index if orientation else 7 - from_file_index) * piece_size + empty_space + light_border_size + dark_border_size
                from_y = (7 - from_rank_index if orientation else from_rank_index) * piece_size + empty_space + light_border_size + dark_border_size

                to_x = (to_file_index if orientation else 7 - to_file_index) * piece_size + empty_space + light_border_size + dark_border_size
                to_y = (7 - to_rank_index if orientation else to_rank_index) * piece_size + empty_space + light_border_size + dark_border_size

                pygame.draw.arrow(
                    self.screen,
                    THECOLORS[_arrow.color],
                    (from_x + piece_size // 2, from_y + piece_size // 2),
                    (to_x + piece_size // 2, to_y + piece_size // 2),
                    5
                )

            pygame.display.flip()
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    exit(1)
        else:
            print(self.unicode())



import pygame

from joatmon.event import Event
from joatmon.system.gui.widgets.label import Label
from joatmon.system.gui.widgets.panel import Panel
from joatmon.system.gui.widgets.widget import Element


class TabView(Element):
    def __init__(self, id_):
        super(TabView, self).__init__(id_)

        self._content = Panel(f'{id_}.content').layout('vertical')
        self._header = Panel(f'{id_}.header').layout('horizontal').style({'height': 1, 'border': {'color': (0, 0, 0), 'thickness': 1, 'shadow': 1, 'radius': 1}})
        self._body = Panel(f'{id_}.body').layout('vertical').style({'height': 9, 'border': {'color': (0, 0, 0), 'thickness': 1, 'shadow': 1, 'radius': 1}})

        self._content.add(self._header)
        self._content.add(self._body)

        self._bodies = {}

        self.add(self._content)

        self._selected_tab = None
        self._events['tab_change'] = Event()

    def _tab_changed(self, tab):
        self._events.get('tab_change').fire(tab)
        self._selected_tab = tab
        self._body._elements = self._bodies[self._selected_tab._id]._elements

        self.update()

    def new_tab(self, tab):
        t = Label(f'{tab}').text(tab).style({'border': {'color': (0, 0, 0), 'thickness': 1, 'shadow': 1, 'radius': 1}}).on('click', self._tab_changed)
        self._header.add(t)

        body = Panel('').layout('vertical').style({'height': 9, 'border': {'color': (0, 0, 0), 'thickness': 1, 'shadow': 1, 'radius': 1}})

        self._bodies[tab] = body

        if self._selected_tab is None:
            self._tab_changed(t)

        return self

    def update_tab(self, tab, *content):
        if tab not in self._bodies:
            self.new_tab(tab)

        body = self._bodies[tab]
        body._elements = []

        for row in content:
            self.new_row(tab, row)

        if self._selected_tab._id == tab:
            self._body._elements = body._elements

        return self

    def get_tab(self, tab):
        body = self._bodies[tab]
        return body

    def new_row(self, tab, text):
        body = self._bodies[tab]
        body.add(Label('').text(text).style({'border': {'color': (0, 0, 0), 'thickness': 1, 'shadow': 1, 'radius': 1}}))
        self._bodies[tab] = body

        return self

    def draw(self, screen):
        self.update()

        super(TabView, self).draw(screen)

        if self._selected_tab:
            pygame.draw.rect(screen, (222, 222, 222), self._selected_tab._rect)



import json
import threading
import time
from threading import Thread

from joatmon import context
from joatmon.event import Event
from joatmon.utility import JSONEncoder

__all__ = ['producer', 'consumer']


def producer(kafka, topic):
    def _decorator(func):
        def _wrapper(*args, **kwargs):
            p = context.get_value(kafka).get_producer(topic)
            message = json.dumps({'args': args, 'kwargs': kwargs}, cls=JSONEncoder).encode('utf-8')
            p.produce(message)
            return func(*args, **kwargs)

        return _wrapper

    return _decorator


consumers = {}
consumer_threads = {}
consumer_events = {}


def loop(topic, cons):
    for message in cons:
        time.sleep(0.1)

        if not threading.main_thread().is_alive():
            break

        if message is None:
            continue

        cons.commit_offsets()

        packet = json.loads(message.value.decode('utf-8'))
        args = packet['args']
        kwargs = packet['kwargs']

        consumer_events[topic].fire(*args, **kwargs)


def consumer_loop_creator():
    global consumer_threads
    while threading.main_thread().is_alive():
        consumer_threads = {topic: thread for topic, thread in consumer_threads.items() if thread.is_alive()}

        for topic, cons in consumers.items():
            if topic not in consumer_threads:
                thread = Thread(target=loop, args=(topic, cons))
                consumer_threads[topic] = thread
                thread.start()
        time.sleep(0.1)


def add_consumer(topic, c):
    if topic not in consumers:
        consumers[topic] = c
        consumer_events[topic] = Event()


# need to have another parameter called is_batch or batch_size
# if False or None consumer will be working with one message at a time
# if not False and None consumer will be working in batch mode
def consumer(kafka, topic):
    def _decorator(func):
        c = context.get_value(kafka).get_consumer(topic)

        add_consumer(topic, c)
        consumer_events[topic] += func

        return func

    return _decorator


Thread(target=consumer_loop_creator).start()



import numpy as np


class CoreRandom(object):
    """
    Abstract base class for all implemented random processes.

    Do not use this abstract base class directly but instead
    use one of the concrete random processes implemented.

    To implement your own random processes,
    you have to implement the following methods:

    - `decay`
    - `sample`
    - `reset`
    """

    def __init__(self):
        super(CoreRandom, self).__init__()

    def reset(self):
        """
        Reset random state.
        """
        raise NotImplementedError

    def decay(self):
        """
        decay
        """
        raise NotImplementedError

    def sample(self):
        """
        Sample random state.

        # Returns
            sample (abstract): Random state.
        """
        raise NotImplementedError


# another class to make the noise decayed, parent class for all
class GaussianRandom(CoreRandom):
    """
    Gaussian Noise

    # Arguments
        mu (float): .
        size (int): .
        sigma (float): .
        sigma_min (float): .
        decay_steps (int): .
    """

    def __init__(self, mu=0.0, size=2, sigma=0.1, sigma_min=0.01, decay_steps=200000):
        super().__init__()

        self.mu = mu
        self.sigma = sigma
        self.sigma_min = sigma_min
        self.size = size

        self.m = -float(sigma - sigma_min) / float(decay_steps)
        self.c = sigma

        self.n_steps = 0
        self.current_sigma = max(self.sigma_min, self.m * float(self.n_steps) + self.c)
        self.reset()

    def reset(self):
        self.n_steps = 0
        self.current_sigma = max(self.sigma_min, self.m * float(self.n_steps) + self.c)

    def decay(self):
        self.n_steps += 1
        self.current_sigma = max(self.sigma_min, self.m * float(self.n_steps) + self.c)

    def sample(self):
        x = np.random.normal(loc=self.mu, scale=self.current_sigma, size=self.size)
        return x


class OrnsteinUhlenbeck(CoreRandom):
    """
    Ornstein Uhlenbeck Process

    # Arguments
        dt (float): .
        mu (float): .
        size (int): .
        sigma (float): .
        theta (float): .
        sigma_min (float): .
        decay_steps (int): .
    """

    def __init__(self, dt=1.0, mu=0.0, size=2, sigma=0.1, theta=0.15, sigma_min=0.01, decay_steps=200000):
        super().__init__()

        self.dt = dt
        self.mu = mu
        self.size = size
        self.sigma = sigma
        self.theta = theta
        self.sigma_min = sigma_min

        self.m = -float(sigma - sigma_min) / float(decay_steps)
        self.c = sigma

        self.n_steps = 0
        self.x_prev = np.ones(self.size) * self.mu
        self.current_sigma = max(self.sigma_min, self.m * float(self.n_steps) + self.c)
        self.reset()

    def reset(self):
        self.n_steps = 0
        self.x_prev = np.ones(self.size) * self.mu
        self.current_sigma = max(self.sigma_min, self.m * float(self.n_steps) + self.c)

    def decay(self):
        self.n_steps += 1
        self.current_sigma = max(self.sigma_min, self.m * float(self.n_steps) + self.c)

    def sample(self):
        x = self.x_prev + self.theta * (self.mu - self.x_prev) * self.dt + self.current_sigma * np.sqrt(self.dt) * np.random.normal(size=self.size)
        self.x_prev = x
        return x



from collections import deque


def preorder(root):
    if root is None:
        return

    stack = deque()
    stack.append(root)

    while stack:
        curr = stack.pop()

        yield curr.data
        if curr.right:
            stack.append(curr.right)
        if curr.left:
            stack.append(curr.left)


def preorder_r(root):
    if root is None:
        return

    yield root.data
    preorder_r(root.left)
    preorder_r(root.right)



import numpy as np

from joatmon.nn import functional as f
from joatmon.nn.core import (
    Module,
    Parameter,
    Tensor
)

__all__ = ['Linear']


class Linear(Module):
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: Tensor

    def __init__(self, in_features: int, out_features: int, bias: bool = True) -> None:
        super(Linear, self).__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(Tensor.from_array(data=np.ones((self.out_features, self.in_features))))
        if bias:
            self.bias = Parameter(Tensor.from_array(data=np.zeros((self.out_features,))))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        self.weight = Parameter(Tensor.from_array(data=np.ones((self.out_features, self.in_features))))
        if self.bias is not None:
            self.bias = Parameter(Tensor.from_array(data=np.zeros((self.out_features,))))

    def forward(self, inp):
        x = f.dense(
            inp=inp,
            weight=self.weight,
            bias=self.bias
        )
        return x

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, bias={}'.format(self.in_features, self.out_features, self.bias is not None)



import torch
import torch.nn as nn

__all__ = ['DDPGActor', 'DDPGCritic']


class DDPGActor(nn.Module):
    def __init__(self, in_features, out_features):
        super(DDPGActor, self).__init__()

        self.extractor = nn.Sequential(
            nn.Conv2d(in_channels=in_features, out_channels=32, kernel_size=(8, 8), stride=(4, 4)),
            nn.BatchNorm2d(num_features=32),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(4, 4), stride=(2, 2)),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=(3, 3), stride=(1, 1)),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(inplace=True)
        )

        self.predictor = nn.Sequential(
            nn.Linear(in_features=7 * 7 * 64, out_features=200),
            nn.BatchNorm1d(num_features=200),
            nn.ReLU(inplace=True),
            nn.Linear(in_features=200, out_features=200),
            nn.BatchNorm1d(num_features=200),
            nn.ReLU(inplace=True),
            nn.Linear(in_features=200, out_features=out_features),
            nn.Tanh()
        )

    def forward(self, x):
        x = self.extractor(x)
        return self.predictor(x.view(x.size(0), -1))


class DDPGCritic(nn.Module):
    def __init__(self, in_features, out_features):
        super(DDPGCritic, self).__init__()

        self.extractor = nn.Sequential(
            nn.Conv2d(in_channels=in_features, out_channels=32, kernel_size=(8, 8), stride=(4, 4)),
            nn.BatchNorm2d(num_features=32),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(4, 4), stride=(2, 2)),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=(3, 3), stride=(1, 1)),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(inplace=True)
        )

        self.relu = nn.ReLU(inplace=True)

        self.linear1 = nn.Linear(in_features=7 * 7 * 64, out_features=200)
        self.linear2 = nn.Linear(in_features=200 + out_features, out_features=200)
        self.linear3 = nn.Linear(in_features=200, out_features=1)

        self.bn1 = nn.BatchNorm1d(200)
        self.bn2 = nn.BatchNorm1d(200)

    def forward(self, x, y):
        x = self.extractor(x)

        x = self.relu(self.bn1(self.linear1(x.view(x.size(0), -1))))
        x = torch.cat([x, y], dim=1)
        x = self.relu(self.bn2(self.linear2(x)))
        return self.linear3(x)



from collections import deque


def postorder(root):
    if root is None:
        return

    stack = deque()
    stack.append(root)
    out = deque()

    while stack:
        curr = stack.pop()
        out.append(curr.data)
        if curr.left:
            stack.append(curr.left)
        if curr.right:
            stack.append(curr.right)

    while out:
        yield out.pop()


def postorder_r(root):
    if root is None:
        return

    postorder_r(root.left)
    postorder_r(root.right)
    yield root.data



import re
import typing
import uuid
from datetime import datetime
from time import (
    mktime,
    struct_time
)

from joatmon.serializable import Serializable
from joatmon.utility import to_enumerable


class Field(Serializable):
    def __init__(
            self,
            dtype: typing.Union[type, typing.List, typing.Tuple],
            nullable: bool = True,
            default=None,
            primary: bool = False,
            encrypt: bool = False,
            hash_: bool = False,
            resource: str = None,
            regex: str = None,
            fields: dict = None
    ):
        super(Field, self).__init__()

        self.dtype = dtype
        self.nullable = nullable
        self.primary = primary
        self.encrypt = encrypt
        self.hash_ = hash_
        self.resource = resource
        self.regex = regex
        self.fields = fields or {}

        # is resource template changed

        if not callable(default):
            self.default = lambda: default
        else:
            self.default = default

        self.encrypted = False
        self.hashed = False


def get_converter(field: Field):
    def _datetime_converter(value: object) -> typing.Optional[datetime]:
        if value is None:
            return value

        if isinstance(value, datetime):
            return value

        if isinstance(value, struct_time):
            return datetime.fromtimestamp(mktime(value))

        if isinstance(value, str):
            return datetime.fromisoformat(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to datetime')

    def _integer_converter(value: object) -> typing.Optional[int]:
        if value is None:
            return value

        if isinstance(value, float):
            return int(value)

        if isinstance(value, int):
            return int(value)

        if isinstance(value, str):
            return int(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to integer')

    def _float_converter(value: object) -> typing.Optional[float]:
        if value is None:
            return value

        if isinstance(value, float):
            return value

        if isinstance(value, int):
            return float(value)

        if isinstance(value, str):
            return float(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to float')

    def _string_converter(value: object) -> typing.Optional[str]:
        if value is None:
            return value

        if isinstance(value, str):
            # print(re.findall(r'{@resource\.(.*?)}', value))
            if field.resource is not None and len(re.findall(field.regex, value)) == 0:
                value = field.resource.format(value)
            return value

        # need to convert bytes as well

        raise ValueError(f'cannot convert {value} with type {type(value)} to string')

    def _byte_converter(value: object) -> typing.Optional[bytes]:
        if value is None:
            return value

        if isinstance(value, bytes):
            return value

        # need to convert string and int list as well

        raise ValueError(f'cannot convert {value} with type {type(value)} to bytes')

    def _boolean_converter(value: object) -> typing.Optional[bool]:
        if value is None:
            return value

        if isinstance(value, bool):
            return value

        # need to convert float, int and str as well

        raise ValueError(f'cannot convert {value} with type {type(value)} to boolean')

    def _uuid_converter(value: object) -> typing.Optional[uuid.UUID]:
        if value is None:
            return value

        if isinstance(value, uuid.UUID):
            return value

        if isinstance(value, str):
            return uuid.UUID(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to uuid')

    def _dictionary_converter(value: object) -> typing.Optional[dict]:
        if value is None:
            return value

        if isinstance(value, dict):
            return to_enumerable(value)

        if hasattr(value, 'dict'):
            return to_enumerable(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to dictionary')

    def _list_converter(value: object) -> typing.Optional[list]:
        if value is None:
            return value

        # if isinstance(value, str):
        #     value = json.loads(value)

        if isinstance(value, (list, tuple, set)):
            ret = []
            for v in value:
                if field.fields is not None:
                    obj = {}
                    for field_name, field_meta in field.fields.items():
                        obj[field_name] = get_converter(field_meta)(v[field_name])
                    ret.append(obj)
                else:
                    ret.append(to_enumerable(v))
            return ret

        raise ValueError(f'cannot convert {value} with type {type(value)} to list')

    def _tuple_converter(value: object) -> typing.Optional[tuple]:
        if value is None:
            return value

        # if isinstance(value, str):
        #     value = json.loads(value)

        if isinstance(value, (list, tuple, set)):
            ret = []
            for v in value:
                ret.append(to_enumerable(v))
            ret = tuple(ret)
            return ret

        raise ValueError(f'cannot convert {value} with type {type(value)} to tuple')

    def _set_converter(value: object) -> typing.Optional[set]:
        if value is None:
            return value

        # if isinstance(value, str):
        #     value = json.loads(value)

        if isinstance(value, (list, tuple, set)):
            ret = []
            for v in value:
                ret.append(to_enumerable(v))
            ret = set(ret)
            return ret

        raise ValueError(f'cannot convert {value} with type {type(value)} to tuple')

    def _object_converter(value: object) -> typing.Optional[object]:
        if value is None:
            return value

        if isinstance(value, dict):
            if field.fields is not None:
                obj = {}
                for field_name, field_meta in field.fields.items():
                    obj[field_name] = get_converter(field_meta)(value[field_name])
                return _dictionary_converter(obj)
            else:
                return _dictionary_converter(value)

            # return _dictionary_converter(value)

        if isinstance(value, list):
            return _list_converter(value)

        if isinstance(value, tuple):
            return _tuple_converter(value)

        if isinstance(value, str):
            return _string_converter(value)

        if hasattr(value, 'dict'):
            return value.dict

        return value

    converters = {
        datetime: _datetime_converter,
        int: _integer_converter,
        float: _float_converter,
        str: _string_converter,
        bytes: _byte_converter,
        bool: _boolean_converter,
        uuid.UUID: _uuid_converter,
        dict: _dictionary_converter,
        list: _list_converter,
        tuple: _tuple_converter,
        set: _set_converter,
        object: _object_converter
    }

    return converters.get(field.dtype, _object_converter)



import datetime
import importlib.util
import json
import os
import threading
import time

import openai

from joatmon import context
from joatmon.assistant import (
    service,
    task
)
from joatmon.assistant.intent import GenericAssistant
from joatmon.assistant.service import (
    ServiceInfo,
    ServiceState
)
from joatmon.assistant.stt import STTAgent
from joatmon.assistant.task import (
    BaseTask,
    TaskInfo,
    TaskState
)
from joatmon.assistant.tts import TTSAgent
from joatmon.system.hid.console import (
    ConsoleReader,
    ConsoleWriter
)
from joatmon.system.hid.microphone import Microphone
from joatmon.system.hid.speaker import Speaker
from joatmon.system.lock import RWLock
from joatmon.utility import get_module_classes


class CTX:
    ...


ctx = CTX()
context.set_ctx(ctx)

openai.api_key = json.loads(open('iva/iva.json', 'r').read())['config']['openai']['key']


class API:
    def __init__(self):
        settings = json.loads(open('iva/iva.json', 'r').read())

        self.tts = settings.get('config', {}).get('tts', False)
        self.stt = settings.get('config', {}).get('stt', False)

        if self.tts:
            self.tts_agent = TTSAgent()
            self.speaker = Speaker()
        else:
            self.writer = ConsoleWriter()

        if self.stt:
            self.stt_agent = STTAgent()
            self.microphone = Microphone()
        else:
            self.reader = ConsoleReader()

        self.output('input and output devices are initialized')

        self.assistant = GenericAssistant('iva/iva.json')
        self.assistant.train_model()
        self.assistant.save_model('iva')

        self.output('intent assistant is initialized')

        self.parent_os_path = os.path.abspath(os.path.curdir)
        self.os_path = os.sep

        self.lock = RWLock()
        self.running_tasks = {}  # running, finished
        self.running_services = {}  # running, enabled, disabled, stopped, finished

        self.event = threading.Event()

        self.output('running startup tasks')
        tasks = settings.get('tasks', [])
        for _task in sorted(filter(lambda x: x['status'] and x['on'] == 'startup', tasks), key=lambda x: x['priority']):
            self.run_task(_task['name'])  # need to do them in background

        self.output('starting automatic services')
        services = settings.get('services', [])
        for _service in sorted(filter(lambda x: x['status'] and x['mode'] == 'automatic', services), key=lambda x: x['priority']):
            self.start_service(_service['name'])  # need to do them in background

        self.output('creating cleanup thread')
        self.cleaning_thread = threading.Thread(target=self.clean)
        self.cleaning_thread.start()
        time.sleep(1)

        self.output('starting service runner thread')
        self.service_thread = threading.Thread(target=self.run_services)
        self.service_thread.start()
        time.sleep(1)

        self.output('creating task scheduler thread')
        self.interval_thread = threading.Thread(target=self.run_interval)
        self.interval_thread.start()
        time.sleep(1)

        # make it async
        # need event viewer

    def run_interval(self):
        while not self.event.is_set():
            settings = json.loads(open('iva/iva.json', 'r').read())
            tasks = settings.get('tasks', [])
            tasks = filter(lambda x: x['status'], tasks)
            tasks = filter(lambda x: x['on'] == 'interval', tasks)
            tasks = filter(lambda x: x['interval'] > 0, tasks)
            tasks = list(tasks)

            new_tasks = filter(lambda x: x.get('last_run_time', None) is None, tasks)
            new_tasks = list(new_tasks)
            old_tasks = filter(lambda x: x.get('next_run_time', None) is not None, tasks)
            old_tasks = filter(lambda x: datetime.datetime.now() > datetime.datetime.fromisoformat(x['next_run_time']), old_tasks)
            old_tasks = list(old_tasks)

            for _task in sorted(new_tasks, key=lambda x: x['priority']):
                self.run_task(task_name=_task['name'], kwargs=None, background=True)  # need to do them in background
            for _task in sorted(old_tasks, key=lambda x: x['priority']):
                self.run_task(task_name=_task['name'], kwargs=None, background=True)  # need to do them in background

            # need to run to do as well
            # need to run agenda as well

            time.sleep(1)

    def run_services(self):
        settings = json.loads(open('iva/iva.json', 'r').read())

        services = settings.get('services', [])

        # if the service is closed for some reason and is configured as restart automatically, need to restart the service

        while not self.event.is_set():
            for _service in sorted(filter(lambda x: x['status'], services), key=lambda x: x['priority']):
                if _service['name'] not in self.running_services or self.running_services[_service['name']].state == ServiceState.finished:
                    self.start_service(_service['name'])  # need to do them in background
            time.sleep(1)

    def clean(self):
        while not self.event.is_set():
            with self.lock.r_locked():
                task_keys = [key for key in self.running_tasks.keys()]
                service_keys = [key for key in self.running_services.keys()]

            delete_task_keys = []
            for key in task_keys:
                task_info = self.running_tasks[key]
                if not task_info.task.running():
                    delete_task_keys.append(key)
            delete_service_keys = []
            for key in service_keys:
                task_info = self.running_services[key]
                if not task_info.service.running() and task_info.state != ServiceState.stopped:
                    delete_service_keys.append(key)

            for key in delete_task_keys:
                with self.lock.w_locked():
                    del self.running_tasks[key]
            for key in delete_service_keys:
                with self.lock.w_locked():
                    del self.running_services[key]

            time.sleep(1)

    def listen_intent(self):
        response = self.input('what is your intent')
        intent, prob = self.intent(response)
        if prob > .9:
            return intent
        else:
            return response

    def listen_command(self):
        response = self.input('what is your command')

        tasks = []

        settings = json.loads(open('iva/iva.json', 'r').read())
        for scripts_folder in settings.get('scripts', []):
            if os.path.isabs(scripts_folder):
                for script_file in os.listdir(scripts_folder):
                    spec = importlib.util.spec_from_file_location(scripts_folder, os.path.join(scripts_folder, script_file))
                    action_module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(action_module)

                    for class_ in get_module_classes(action_module):
                        if not issubclass(class_[1], BaseTask) or class_[1] is BaseTask:
                            continue

                        tasks.append(class_[1])
            else:
                try:
                    _module = __import__('.'.join(scripts_folder.split('.')), fromlist=[scripts_folder.split('.')[-1]])

                    scripts_folder = _module.__path__[0]

                    for script_file in os.listdir(scripts_folder):
                        if '__' in script_file:
                            continue

                        spec = importlib.util.spec_from_file_location(scripts_folder, os.path.join(scripts_folder, script_file))
                        action_module = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(action_module)

                        for class_ in get_module_classes(action_module):
                            if not issubclass(class_[1], BaseTask) or class_[1] is BaseTask:
                                continue

                            tasks.append(class_[1])
                except ModuleNotFoundError:
                    print('module not found')
                    continue

        functions = list(map(lambda x: x.help(), tasks))
        functions = list(filter(lambda x: x, functions))

        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo-0613",
            messages=[{"role": "user", "content": response}],
            functions=functions
        )

        result = response['choices'][0]
        if result['finish_reason'] == 'function_call':
            message = result['message']

            return message['function_call']['name'], json.loads(message['function_call']['arguments'])

    def input(self, prompt=None):
        # put the input to the queue
        # if the task is called from outside, putting inputs to the queue by hand will make it work

        # if not english, translate to english
        # then get the response
        # get the intent
        # run the action
        # translate the response if needed

        if prompt:
            self.output(prompt)

        if self.stt:
            response = self.microphone.listen()
            response = self.stt_agent.transcribe(response)
        else:
            response = self.reader.read()
        return response

    def output(self, text):
        if self.tts:
            text = self.tts_agent.convert(text)
            self.speaker.say(text)
        else:
            self.writer.write(text)

    def intent(self, text):
        return self.assistant.request(text)

    def do_action(self, line):
        try:
            if line is None or line == '':
                return False

            match line.lower():
                case 'enable':
                    return False
                case 'disable':
                    return False
                case 'update':
                    return False
                case 'delete':
                    return False
                case 'configure':
                    return False
                case 'start':
                    return False
                case 'stop':
                    return False
                case 'restart':
                    return False
                case 'skip':
                    return False
                case 'help':
                    return False
                case 'exit':
                    return self.exit()
                case 'activate':
                    command, arguments = self.listen_command()
                    return self.run_task(task_name=command, kwargs=arguments)
                case _:
                    raise ValueError(f'wanted command is {line}, default case is not implemented')
        except Exception as ex:
            print(str(ex))  # use stacktrace and write all exception details, line number, function name, file name etc.
            # return self.exit()

    def run_task(self, task_name, kwargs=None, background=False):
        _task = task.get(self, task_name, kwargs, background)
        if _task is None:
            return False

        if _task.background:
            if task_name not in self.running_tasks:
                self.running_tasks[task_name] = TaskInfo(task_name, TaskState.running, _task)
            else:
                self.running_tasks[task_name].state = TaskState.running
                self.running_tasks[task_name].task = _task
        _task.start()

        return False

    def start_service(self, service_name):
        _task = service.get(self, service_name)
        if _task is None:
            return False

        if service_name not in self.running_services:
            self.running_services[service_name] = ServiceInfo(service_name, ServiceState.running, _task)
        else:
            self.running_services[service_name].state = ServiceState.running
            self.running_services[service_name].service = _task
        _task.start()

        return False

    def stop_service(self, service_name):
        self.running_services[service_name].state = ServiceState.stopped
        self.running_services[service_name].service.stop()
        return False

    def restart_service(self, service_name):
        self.stop_service(service_name)
        self.start_service(service_name)
        return False

    def exit(self):
        self.output('shutting down the system')

        settings = json.loads(open('iva/iva.json', 'r').read())

        self.output('running shutdown tasks')
        tasks = settings.get('tasks', [])
        for _task in sorted(filter(lambda x: x['status'] and x['on'] == 'shutdown', tasks), key=lambda x: x['priority']):
            self.run_task(_task['name'])

        self.output('stopping background tasks, jobs and services')
        with self.lock.r_locked():
            task_keys = [key for key in self.running_tasks.keys()]
            service_keys = [key for key in self.running_services.keys()]

        for key in task_keys:
            task_info = self.running_tasks[key]
            task_info.task.stop()
        for key in service_keys:
            self.stop_service(key)

        self.event.set()
        self.output('closing input device')
        if self.stt:
            self.microphone.stop()
        self.output('shutdown')
        return True

    def mainloop(self):
        while not self.event.is_set():
            command = self.listen_intent()
            self.do_action(command)


def main():
    api = API()
    api.mainloop()



from __future__ import print_function

import os

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "save",
            "description": "a function for user to save a message to a file",
            "parameters": {
                "type": "object",
                "properties": {
                    "message": {
                        "type": "string",
                        "description": "content of the file"
                    },
                    "path": {
                        "type": "string",
                        "description": "path of the file"
                    }
                },
                "required": ["message", "path"]
            }
        }

    def run(self):
        path = self.kwargs.get('path', '') or self.api.input('what is the path')
        message = self.kwargs.get('message', '') or self.api.input('what is the message')

        parent_os_path = self.kwargs.get('parent_os_path', '')
        os_path = self.kwargs.get('os_path', '')

        p = path if os.path.isabs(path) else os.path.join(parent_os_path, os_path[1:], path)

        with open(p, 'wb') as file:
            file.write(message.encode('utf-8', errors='ignore'))

        if not self.stop_event.is_set():
            self.stop_event.set()



import copy
import os

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from joatmon.ai.models.core import CoreModel
from joatmon.ai.network import DQN
from joatmon.ai.utility import (
    load,
    range_tensor,
    save,
    to_numpy,
    to_tensor
)

__all__ = ['DQNModel']


class DQNModel(CoreModel):
    """
    Deep Q Network

    # Arguments
        models (`keras.nn.Model` instance): See [Model](#) for details.
        optimizer (`keras.optimizers.Optimizer` instance):
        See [Optimizer](#) for details.
        tau (float): tau.
        gamma (float): gamma.
    """

    def __init__(self, lr=1e-3, tau=1e-4, gamma=0.99, in_features=1, out_features=1):
        super().__init__()

        self.local = DQN(in_features, out_features)
        self.target = copy.deepcopy(self.local)
        self.hardupdate()

        self.lr = lr
        self.tau = tau
        self.gamma = gamma

        self.optimizer = optim.Adam(self.local.parameters(), lr=self.lr)
        self.loss = nn.SmoothL1Loss()

    def load(self, path=''):
        load(self.local, os.path.join(path, 'local'))
        load(self.target, os.path.join(path, 'target'))

    def save(self, path=''):
        save(self.local, os.path.join(path, 'local'))
        save(self.target, os.path.join(path, 'target'))

    def initialize(self, w_init=None, b_init=None):
        for module in self.target.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)
        for module in self.local.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)

    def softupdate(self):
        for target_param, param in zip(self.target.parameters(), self.local.parameters()):
            target_param.detach_()
            target_param.copy_(target_param * (1.0 - self.tau) + param * self.tau)

    def hardupdate(self):
        for target_param, param in zip(self.target.parameters(), self.local.parameters()):
            target_param.detach_()
            target_param.copy_(param)

    def predict(self, state=None):
        self.local.eval()
        return np.argmax(to_numpy(self.local(to_tensor(state))).flatten())

    def train(self, batch=None, update_target=False):
        self.local.train()

        states, actions, rewards, next_states, terminals = batch

        states = to_tensor(states)
        actions = to_tensor(actions)
        rewards = to_tensor(rewards)
        next_states = to_tensor(next_states)
        terminals = to_tensor(terminals)

        batch_indices = range_tensor(states.size(0))

        q_next = self.target(next_states).detach()
        q_next = q_next.max(1)[0]
        q_next = self.gamma * q_next * (1 - terminals)
        q_next.add_(rewards)

        q = self.local(states)
        q = q[batch_indices, actions.long()]
        loss = self.loss(q, q_next)
        self.optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(self.local.parameters(), 1)
        self.optimizer.step()

        if update_target:
            self.softupdate()

        return loss.item()

    def evaluate(self):
        pass
